from typing import Optional

import torch
import torchvision.transforms as T
from einops import rearrange

from torchrs.datasets.utils import dataset_split
from torchrs.train.datamodules import BaseDataModule
from torchrs.transforms import ToTensor, ToDtype
from torchrs.datasets import PROBAV


class PROBAVDataModule(BaseDataModule):

    def __init__(
        self,
        root: str = ".data/probav",
        band: str = "RED",
        lr_transform: T.Compose = T.Compose([ToTensor(), ToDtype(torch.float32)]),
        hr_transform: T.Compose = T.Compose([ToTensor(), ToDtype(torch.float32)]),
        *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.root = root
        self.band = band
        self.lr_transform = lr_transform
        self.hr_transform = hr_transform

    def setup(self, stage: Optional[str] = None):
        train_dataset = PROBAV(
            root=self.root, split="train", band=self.band, lr_transform=self.lr_transform, hr_transform=self.hr_transform
        )
        self.train_dataset, self.val_dataset = dataset_split(train_dataset, val_pct=self.val_split)
        self.test_dataset = PROBAV(
            root=self.root, split="test", band=self.band, lr_transform=self.lr_transform, hr_transform=self.hr_transform
        )

    def on_before_batch_transfer(self, batch, dataloader_idx):
        """ Handle if lr and hr are chipped by ExtractChips transform """
        if batch["lr"].ndim == 6:
            batch["lr"] = rearrange(batch["lr"], "b t d c h w -> (b d) t c h w")
        if batch["hr"].ndim == 5:
            batch["hr"] = rearrange(batch["hr"], "b d c h w -> (b d) c h w")
        return batch



import torchvision.transforms as T
from torchvision.datasets import ImageFolder


class WHURS19(ImageFolder):
    """ WHU-RS19 dataset from'Structural High-resolution Satellite Image Indexing', Xia at al. (2010)
    https://hal.archives-ouvertes.fr/file/index/docid/458685/filename/structural_satellite_indexing_XYDG.pdf

    """
    def __init__(
        self,
        root: str = ".data/WHU-RS19",
        transform: T.Compose = T.Compose([T.ToTensor()])
    ):
        super().__init__(
            root=root,
            transform=transform
        )



import os
from glob import glob
from typing import Dict

import torch
import tifffile
import numpy as np
from PIL import Image

from torchrs.transforms import Compose, ToTensor


def sort(x):
    band = os.path.splitext(x.split(os.sep)[-1])[0]
    if band == "B8A":
        band = "B08A"
    return band


class OSCD(torch.utils.data.Dataset):
    """ The Onera Satellite Change Detection (OSCD) dataset from 'Urban Change Detection for
    Multispectral Earth Observation Using Convolutional Neural Networks', Daudt at al. (2018)
    https://arxiv.org/abs/1703.00121

    'The Onera Satellite Change Detection dataset addresses the issue of detecting changes between
    satellite images from different dates. It comprises 24 pairs of multispectral images taken
    from the Sentinel-2 satellites between 2015 and 2018. Locations are picked all over the world,
    in Brazil, USA, Europe, Middle-East and Asia. For each location, registered pairs of 13-band
    multispectral satellite images obtained by the Sentinel-2 satellites are provided. Images vary
    in spatial resolution between 10m, 20m and 60m. Pixel-level change ground truth is provided for
    all 14 training and 10 test image pairs. The annotated changes focus on urban changes, such as
    new buildings or new roads. These data can be used for training and setting parameters of change
    detection algorithms.
    """
    splits = ["train", "test"]

    def __init__(
        self,
        root: str = ".data/oscd",
        split: str = "train",
        transform: Compose = Compose([ToTensor(permute_dims=False)]),
    ):
        assert split in self.splits
        self.root = root
        self.transform = transform
        self.regions = self.load_files(root, split)

    @staticmethod
    def load_files(root: str, split: str):
        regions = []
        labels_root = os.path.join(root, f"{split}_labels")
        images_root = os.path.join(root, "images")
        folders = glob(os.path.join(labels_root, "*/"))
        for folder in folders:
            region = folder.split(os.sep)[-2]
            mask = os.path.join(labels_root, region, "cm", "cm.png")
            images1 = glob(os.path.join(images_root, region, "imgs_1_rect", "*.tif"))
            images2 = glob(os.path.join(images_root, region, "imgs_2_rect", "*.tif"))
            images1 = sorted(images1, key=sort)
            images2 = sorted(images2, key=sort)
            with open(os.path.join(images_root, region, "dates.txt")) as f:
                dates = tuple([line.split()[-1] for line in f.read().strip().splitlines()])

            regions.append(dict(region=region, images1=images1, images2=images2, mask=mask, dates=dates))

        return regions

    def __len__(self) -> int:
        return len(self.regions)

    def __getitem__(self, idx: int) -> Dict:
        """ Returns a dict containing x, mask
        x: (2, 13, h, w)
        mask: (1, h, w)
        """
        region = self.regions[idx]
        mask = np.array(Image.open(region["mask"]))
        mask[mask == 255] = 1
        image1 = np.stack([tifffile.imread(path) for path in region["images1"]], axis=0)
        image2 = np.stack([tifffile.imread(path) for path in region["images2"]], axis=0)
        image1, image2, mask = self.transform([image1, image2, mask])
        x = torch.stack([image1, image2], dim=0)
        return dict(x=x, mask=mask)



from typing import Optional

from torchrs.train.datamodules import BaseDataModule
from torchrs.datasets import S2Looking
from torchrs.transforms import Compose, ToTensor


class S2LookingDataModule(BaseDataModule):

    def __init__(
        self,
        root: str = ".data/s2looking",
        transform: Compose = Compose([ToTensor()]),
        *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.root = root
        self.transform = transform

    def setup(self, stage: Optional[str] = None):
        self.train_dataset = S2Looking(root=self.root, split="train", transform=self.transform)
        self.val_dataset = S2Looking(root=self.root, split="val", transform=self.transform)
        self.test_dataset = S2Looking(root=self.root, split="test", transform=self.transform)



import pytest
import torch
import pytorch_lightning as pl

from torchrs.train import datamodules


skip = ["BaseDataModule", "RSVQAxBENDataModule", "S2MTCPDataModule"]


@torch.no_grad()
@pytest.mark.parametrize("datamodule", reversed(datamodules.__all__))
def test_datamodules(datamodule: pl.LightningDataModule):

    if datamodule in skip:
        return

    dm = getattr(datamodules, datamodule)()
    dm.setup()
    batch = next(iter(dm.train_dataloader()))
    batch = next(iter(dm.val_dataloader()))
    batch = next(iter(dm.test_dataloader()))



import os
from glob import glob
from typing import List, Dict

import torch
import numpy as np
import torchvision.transforms as T
from PIL import Image

from torchrs.transforms import ToTensor, ToDtype


STATS = {
    "RED": {"min": 864.0, "max": 15312.8, "mean": 0.0587, "std": 0.6386},
    "NIR": {"min": 628.8, "max": 17464.0, "mean": 0.0618, "std": 0.7428}
}


def collate_fn(batch: List[Dict], t: int = 9, shuffle: bool = False) -> Dict[str, torch.Tensor]:
    lrs = [x["lr"] for x in batch]
    qms = [x["qm"] for x in batch]

    # Shuffle lr temporal order
    if shuffle:
        perms = [torch.randperm(lr.shape[0]) for lr in lrs]
        lrs = [lr[perm] for perm, lr in zip(perms, lrs)]
        qms = [qm[perm] for perm, qm in zip(perms, qms)]

    # Repeat some images if less than temporal length t
    for i in range(len(lrs)):
        if lrs[i].shape[0] < t:
            n = t - lrs[i].shape[0]
            lrs[i] = torch.cat([lrs[i], lrs[i][:n]], dim=0)
            qms[i] = torch.cat([qms[i], qms[i][:n]], dim=0)

    # Select t lr images from each set
    lrs = torch.stack([lr[:t] for lr in lrs])
    qms = torch.stack([qm[:t] for qm in qms])

    hr = torch.stack([x["hr"] for x in batch])
    sm = torch.stack([x["sm"] for x in batch])

    return dict(lr=lrs, hr=hr, qm=qms, sm=sm)


class PROBAV(torch.utils.data.Dataset):
    """Multi-image super resolution (MISR) dataset from the PROBA-V Super Resolution Competition
    https://kelvins.esa.int/proba-v-super-resolution/home/

    'We collected satellite data from the PROBA-V mission of the European Space Agency from 74 hand-selected
    regions around the globe at different points in time. The data is composed of radiometrically and geometrically
    corrected Top-Of-Atmosphere (TOA) reflectances for the RED and NIR spectral bands at 300m and 100m resolution
    in Plate Carrée projection. The 300m resolution data is delivered as 128x128 grey-scale pixel images, the 100m
    resolution data as 384x384 grey-scale pixel images. The bit-depth of the images is 14, but they are saved in a
    16-bit .png-format (which makes them look relatively dark if opened in typical image viewers).

    Each image comes with a quality map, indicating which pixels in the image are concealed
    (i.e. clouds, cloud shadows, ice, water, missing, etc) and which should be considered clear. For an image to be
    included in the dataset, at least 75% of its pixels have to be clear for 100m resolution images, and 60% for
    300m resolution images. Each data-point consists of exactly one 100m resolution image and several 300m resolution
    images from the same scene. In total, the dataset contains 1450 scenes, which are split into 1160 scenes for
    training and 290 scenes for testing. On average, each scene comes with 19 different low resolution images and
    always with at least 9. We expect you to submit a 384x384 image for each of the 290 test-scenes, for which we
    will not provide a high resolution image.'
    """
    splits = ["train", "test"]

    def __init__(
        self,
        root: str = ".data/probav",
        split: str = "train",
        band: str = "RED",
        lr_transform: T.Compose = T.Compose([ToTensor(), ToDtype(torch.float32)]),
        hr_transform: T.Compose = T.Compose([ToTensor(), ToDtype(torch.float32)]),
    ):
        assert split in self.splits
        assert band in ["RED", "NIR"]
        self.split = split
        self.lr_transform = lr_transform
        self.hr_transform = hr_transform
        self.imgsets = self.load_files(root, split, band)

    @staticmethod
    def load_files(root: str, split: str, band: str) -> List[Dict]:
        imgsets = []
        folders = sorted(glob(os.path.join(root, split, band, "imgset*")))
        for folder in folders:
            lr = sorted(glob(os.path.join(folder, "LR*.png")))
            qm = sorted(glob(os.path.join(folder, "QM*.png")))
            sm = glob(os.path.join(folder, "SM.png"))[0]

            if split == "train":
                hr = glob(os.path.join(folder, "HR.png"))[0]
            else:
                hr = None

            imgsets.append(dict(lr=lr, qm=qm, hr=hr, sm=sm))
        return imgsets

    def __len__(self) -> int:
        return len(self.imgsets)

    def __getitem__(self, idx: int) -> Dict[str, torch.Tensor]:
        """ Returns a dict containing lr, qm, hr, sm
        lr: (t, 1, h, w) low resolution images
        qm: (t, 1, h, w) low resolution image quality masks
        hr: (1, h, w) high resolution image
        sm: (1, h, w) high resolution image status mask

        Note:
        lr/qm original size is (128, 128),
        hr/sm original size is (384, 384) (scale factor = 3)
        t is the number of lr images for an image set (min = 9)
        """
        imgset = self.imgsets[idx]

        # Load
        lrs = [np.array(Image.open(lr), dtype="int32") for lr in imgset["lr"]]
        qms = [np.array(Image.open(qm), dtype="bool") for qm in imgset["qm"]]
        sm = np.array(Image.open(imgset["sm"]), dtype="bool")

        # Transform
        lrs = torch.stack([self.lr_transform(lr) for lr in lrs])
        qms = torch.stack([torch.from_numpy(qm) for qm in qms]).unsqueeze(1)
        sm = torch.from_numpy(sm).unsqueeze(0)

        if self.split == "train":
            hr = np.array(Image.open(imgset["hr"]), dtype="int32")
            hr = self.hr_transform(hr)
            output = dict(lr=lrs, qm=qms, hr=hr, sm=sm)
        else:
            output = dict(lr=lrs, qm=qms, sm=sm)

        return output



import os
from glob import glob
from typing import List, Dict

import torch
import torchaudio
import numpy as np
import torchvision.transforms as T
from PIL import Image


class ADVANCE(torch.utils.data.Dataset):
    """ AuDio Visual Aerial sceNe reCognition datasEt (ADVANCE) from
    'Cross-Task Transfer for Geotagged Audiovisual Aerial Scene Recognition', Hu et al. (2020)
    https://arxiv.org/abs/2005.08449

    'We create an annotated dataset consisting of 5075 geotagged aerial imagesound pairs
    involving 13 scene classes. This dataset covers a large variety of scenes from across
    the world'
    """
    def __init__(
        self,
        root: str = ".data/advance",
        image_transform: T.Compose = T.Compose([T.ToTensor()]),
        audio_transform: T.Compose = T.Compose([]),
    ):
        self.root = root
        self.image_transform = image_transform
        self.audio_transform = audio_transform
        self.files = self.load_files(root)
        self.classes = sorted(set(f["cls"] for f in self.files))

    @staticmethod
    def load_files(root: str) -> List[Dict]:
        images = sorted(glob(os.path.join(root, "vision", "**", "*.jpg")))
        wavs = sorted(glob(os.path.join(root, "sound", "**", "*.wav")))
        labels = [image.split(os.sep)[-2] for image in images]
        files = [dict(image=image, audio=wav, cls=label) for image, wav, label in zip(images, wavs, labels)]
        return files

    def __len__(self) -> int:
        return len(self.files)

    def __getitem__(self, idx: int) -> Dict:
        """ Returns a dict containing image, audio, and class label
        image: (3, 512, 512)
        audio: (1, 220500)
        cls: int
        """
        files = self.files[idx]
        image = np.array(Image.open(files["image"]).convert("RGB"))
        audio, fs = torchaudio.load(files["audio"])
        image = self.image_transform(image)
        audio = self.audio_transform(audio)
        return dict(image=image, audio=audio, cls=files["cls"])



from typing import Optional, Callable

import pytorch_lightning as pl
from torch.utils.data import DataLoader


class BaseDataModule(pl.LightningDataModule):

    def __init__(
        self,
        batch_size: int = 1,
        num_workers: int = 0,
        prefetch_factor: int = 2,
        pin_memory: bool = False,
        collate_fn: Optional[Callable] = None,
        test_collate_fn: Optional[Callable] = None,
        val_split: float = 0.1,
        test_split: float = 0.25
    ):
        super().__init__()
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.prefetch_factor = prefetch_factor
        self.pin_memory = pin_memory
        self.collate_fn = collate_fn
        self.test_collate_fn = test_collate_fn
        self.val_split = val_split
        self.test_split = test_split

    def train_dataloader(self) -> DataLoader:
        return DataLoader(
            self.train_dataset,
            shuffle=True,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            prefetch_factor=self.prefetch_factor,
            pin_memory=self.pin_memory,
            collate_fn=self.collate_fn
        )

    def val_dataloader(self) -> DataLoader:
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            prefetch_factor=self.prefetch_factor,
            pin_memory=self.pin_memory,
            collate_fn=self.test_collate_fn
        )

    def test_dataloader(self) -> DataLoader:
        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            prefetch_factor=self.prefetch_factor,
            pin_memory=self.pin_memory,
            collate_fn=self.test_collate_fn
        )



import os

import tifffile
import torchvision.transforms as T
from torchvision.datasets import ImageFolder

from torchrs.transforms import ToTensor


class EuroSATRGB(ImageFolder):
    """ Sentinel-2 RGB Land Cover Classification dataset from 'EuroSAT: A Novel Dataset
    and Deep Learning Benchmark for Land Use and Land Cover Classification', Helber at al. (2017)
    https://arxiv.org/abs/1709.00029

    'We present a novel dataset based on Sentinel-2 satellite images covering 13 spectral
    bands and consisting out of 10 classes with in total 27,000 labeled and geo-referenced images.'

    Note: RGB bands only
    """
    def __init__(
        self,
        root: str = ".data/eurosat-rgb",
        transform: T.Compose = T.Compose([T.ToTensor()])
    ):
        super().__init__(
            root=os.path.join(root, "2750"),
            transform=transform
        )


class EuroSATMS(ImageFolder):
    """ Sentinel-2 RGB Land Cover Classification dataset from 'EuroSAT: A Novel Dataset
    and Deep Learning Benchmark for Land Use and Land Cover Classification', Helber at al. (2017)
    https://arxiv.org/abs/1709.00029

    'We present a novel dataset based on Sentinel-2 satellite images covering 13 spectral
    bands and consisting out of 10 classes with in total 27,000 labeled and geo-referenced images.'

    Note: all 13 multispectral (MS) bands
    """
    def __init__(
        self,
        root: str = ".data/eurosat-ms",
        transform: T.Compose = T.Compose([ToTensor()])
    ):
        super().__init__(
            root=os.path.join(root, "ds/images/remote_sensing/otherDatasets/sentinel_2/tif"),
            transform=transform,
            loader=tifffile.imread
        )



import torchvision.transforms as T
from torchvision.datasets import ImageFolder


class AID(ImageFolder):
    """ Aerial Image Dataset (AID) from 'AID: A Benchmark Dataset for Performance
    Evaluation of Aerial Scene Classification', Xia et al. (2017)
    https://arxiv.org/abs/1608.05167

    'The AID dataset has a number of 10000 images within 30 classes.'
    """
    def __init__(
        self,
        root: str = ".data/AID",
        transform: T.Compose = T.Compose([T.ToTensor()])
    ):
        super().__init__(
            root=root,
            transform=transform
        )



import os
from collections import namedtuple
from typing import Dict

import h5py
import torch

from torchrs.transforms import Compose, ToTensor


class ZueriCrop(torch.utils.data.Dataset):
    """ ZueriCrop dataset from 'Crop mapping from image time series:
    deep learning with multi-scale label hierarchies', Turkoglu et al. (2021)
    https://arxiv.org/abs/2102.08820

    'We provide a new, publicly available crop classification dataset ZueriCrop, equipped
    with a tree-structured label hierarchy. ZueriCrop covers a 50 km × 48 km area
    in the Swiss cantons of Zurich and Thurgau. It contains 28,000 Sentinel-2 image
    patches of size 24 pixels × 24 pixels, each observed 71 times over a period of
    52 weeks; 48 agricultural land cover classes; and 116,000 individual agricultural fields.'

    """
    ZueriCropClass = namedtuple("ZueriCropClass", ["id", "label", "level1", "level2", "level3", "description"])
    classes = [
        ZueriCropClass(0, "Unknown", "Unknown", "Unknown", "Unknown", "Unknown"),
        ZueriCropClass(1, "SummerBarley", "Vegetation", "Field crops", "SmallGrainCereal", "summer barley"),
        ZueriCropClass(2, "WinterBarley", "Vegetation", "Field crops", "SmallGrainCereal", "Winter barley"),
        ZueriCropClass(3, "Oat", "Vegetation", "Field crops", "SmallGrainCereal", "oats"),
        ZueriCropClass(4, "Wheat", "Vegetation", "Field crops", "SmallGrainCereal", "triticale"),
        ZueriCropClass(5, "Grain", "Vegetation", "Field crops", "SmallGrainCereal", "mixed fodder cereals"),
        ZueriCropClass(6, "Wheat", "Vegetation", "Field crops", "SmallGrainCereal", "Feed wheat according to the list of varieties swiss granum"),
        ZueriCropClass(7, "Maize", "Vegetation", "Field crops", "LargeGrainCereal", "grain maize"),
        ZueriCropClass(8, "EinkornWheat", "Vegetation", "Field crops", "SmallGrainCereal", "wild emmer, einkorn wheat"),
        ZueriCropClass(9, "SummerWheat", "Vegetation", "Field crops", "SmallGrainCereal", "summer wheat"),
        ZueriCropClass(10, "WinterWheat", "Vegetation", "Field crops", "SmallGrainCereal", "Winter wheat (without forage wheat of the variety list swiss granum),"),
        ZueriCropClass(11, "Rye", "Vegetation", "Field crops", "SmallGrainCereal", "rye"),
        ZueriCropClass(12, "Grain", "Vegetation", "Field crops", "SmallGrainCereal", "mixed cereals for bread production"),
        ZueriCropClass(13, "Spelt", "Vegetation", "Field crops", "SmallGrainCereal", "Spelt"),
        ZueriCropClass(14, "Maize", "Vegetation", "Field crops", "LargeGrainCereal", "maize for seed production (by contract),"),
        ZueriCropClass(15, "Maize", "Vegetation", "Field crops", "LargeGrainCereal", "Mads d'ensilage et mads vert"),
        ZueriCropClass(16, "Sugar_beets", "Vegetation", "Field crops", "BroadLeafRowCrop", "Sugar beets"),
        ZueriCropClass(17, "Beets", "Vegetation", "Field crops", "BroadLeafRowCrop", "fodder beet"),
        ZueriCropClass(18, "Potatoes", "Vegetation", "Field crops", "BroadLeafRowCrop", "potatoes"),
        ZueriCropClass(19, "Potatoes", "Vegetation", "Field crops", "BroadLeafRowCrop", "potatoes for planting (by contract),"),
        ZueriCropClass(20, "SummerRapeseed", "Vegetation", "Field crops", "BroadLeafRowCrop", "summer rapeseed for oil production"),
        ZueriCropClass(21, "WinterRapeseed", "Vegetation", "Field crops", "BroadLeafRowCrop", "Winter rape for cooking oil"),
        ZueriCropClass(22, "Soy", "Vegetation", "Field crops", "BroadLeafRowCrop", "soy"),
        ZueriCropClass(23, "Sunflowers", "Vegetation", "Field crops", "BroadLeafRowCrop", "Sunflowers"),
        ZueriCropClass(24, "Linen", "Vegetation", "Field crops", "BroadLeafRowCrop", "flax"),
        ZueriCropClass(25, "Hemp", "Vegetation", "Field crops", "BroadLeafRowCrop", "hemp"),
        ZueriCropClass(26, "Field bean", "Vegetation", "Field crops", "BroadLeafRowCrop", "Field beans for animal feed"),
        ZueriCropClass(27, "Peas", "Vegetation", "Field crops", "BroadLeafRowCrop", "protein peas for animal fodder"),
        ZueriCropClass(28, "Lupine", "Vegetation", "Field crops", "BroadLeafRowCrop", "lupine for fodder"),
        ZueriCropClass(29, "Pumpkin", "Vegetation", "Field crops", "VegetableCrop", "oil pumpkins"),
        ZueriCropClass(30, "Tobacco", "Vegetation", "Field crops", "BroadLeafRowCrop", "tobacco"),
        ZueriCropClass(31, "Sorghum", "Vegetation", "Field crops", "LargeGrainCereal", "millet"),
        ZueriCropClass(32, "Grain", "Vegetation", "Field crops", "SmallGrainCereal", "Ensiled grain"),
        ZueriCropClass(33, "Linen", "Vegetation", "Field crops", "BroadLeafRowCrop", "false flax (camelina sativa),"),
        ZueriCropClass(34, "Vegetables", "Vegetation", "Field crops", "VegetableCrop", "Annual free-range vegetables, without canned vegetables"),
        ZueriCropClass(35, "Vegetables", "Vegetation", "Field crops", "VegetableCrop", "Ground-canned vegetables"),
        ZueriCropClass(36, "Chicory", "Vegetation", "Field crops", "VegetableCrop", "chicory roots"),
        ZueriCropClass(37, "Buckwheat", "Vegetation", "Field crops", "SmallGrainCereal", "buckwheat"),
        ZueriCropClass(38, "Sorghum", "Vegetation", "Field crops", "LargeGrainCereal", "sorghum"),
        ZueriCropClass(39, "Berries", "Vegetation", "Special crops", "Berries", "Annual berries (e.g., strawberries),"),
        ZueriCropClass(40, "Unknown", "Vegetation", "Special crops", "Unknown", "single-year regrowing ressources (Kenaf and others),"),
        ZueriCropClass(41, "Unknown", "Vegetation", "Special crops", "Unknown", "Annual spice and medicinal plants"),
        ZueriCropClass(42, "Unknown", "Vegetation", "Special crops", "Unknown", "Annual horticultural outdoor crops (flowers, turf etc.),"),
        ZueriCropClass(43, "Biodiversity encouragement area", "Vegetation", "Grassland", "BiodiversityArea", "Conservation headlands"),
        ZueriCropClass(44, "Fallow", "Vegetation", "Special crops", "Fallow", "fallow"),
        ZueriCropClass(45, "Fallow", "Vegetation", "Special crops", "Fallow", "rotational set-aside"),
        ZueriCropClass(46, "Unknown", "Vegetation", "Grassland", "Meadow", "Hem on arable land"),
        ZueriCropClass(47, "Unknown", "Vegetation", "Special crops", "Unknown", "Poppy"),
        ZueriCropClass(48, "Unknown", "Vegetation", "Special crops", "Unknown", "safflower"),
        ZueriCropClass(49, "Unknown", "Vegetation", "Field crops", "CropMix", "lentils"),
        ZueriCropClass(50, "MixedCrop", "Vegetation", "Field crops", "CropMix", "Mixtures of field beans, protein peas and lupines for animal feed with cereals, at least 30% legume content at harvest"),
        ZueriCropClass(51, "Biodiversity encouragement area", "Vegetation", "Grassland", "BiodiversityArea", "Bloom strips for pollinators and other beneficials"),
        ZueriCropClass(52, "Mustard", "Vegetation", "Field crops", "BroadLeafRowCrop", "mustard"),
        ZueriCropClass(53, "WinterRapeseed", "Vegetation", "Field crops", "BroadLeafRowCrop", "Winter rape as a renewable raw material"),
        ZueriCropClass(54, "Sunflowers", "Vegetation", "Field crops", "BroadLeafRowCrop", "Sunflowers as regrowing ressource"),
        ZueriCropClass(55, "Unknown", "Bare soil", "Unknown", "Unknown", "open arable land, eligible for subsidies (region-specific biodiversity area),"),
        ZueriCropClass(56, "Unknown", "Bare soil", "Unknown", "Unknown", "Other open arable land, eligible"),
        ZueriCropClass(57, "Unknown", "Bare soil", "Unknown", "Unknown", "Other open arable land, not eligible"),
        ZueriCropClass(58, "Meadow", "Vegetation", "Grassland", "Meadow", "Art meadows (without pastures),"),
        ZueriCropClass(59, "Meadow", "Vegetation", "Grassland", "Meadow", "Other artificial meadow, eligible (eg pork pasture, poultry pasture),"),
        ZueriCropClass(60, "Meadow", "Vegetation", "Grassland", "Meadow", "Extensively used meadows (without pastures),"),
        ZueriCropClass(61, "Meadow", "Vegetation", "Grassland", "Meadow", "Little intensively used meadows (without pastures),"),
        ZueriCropClass(62, "Meadow", "Vegetation", "Grassland", "Meadow", "Other permanent pastures (without pastures),"),
        ZueriCropClass(63, "Pasture", "Vegetation", "Grassland", "Pasture", "Pastures (pastures, other pastures without summer pastures),"),
        ZueriCropClass(64, "Pasture", "Vegetation", "Grassland", "Pasture", "Extensively used pastures"),
        ZueriCropClass(65, "Pasture", "Vegetation", "Grassland", "Pasture", "Forest pastures (without wooded area),"),
        ZueriCropClass(66, "Meadow", "Vegetation", "Grassland", "Meadow", "Hay meadows in the summering area, other meadows"),
        ZueriCropClass(67, "Meadow", "Vegetation", "Grassland", "Meadow", "Hay meadows in the summering area, type extensively used meadow"),
        ZueriCropClass(68, "Pasture", "Vegetation", "Grassland", "Pasture", "Forest pastures (without wooded area),"),
        ZueriCropClass(69, "Legumes", "Vegetation", "Field crops", "BroadLeafRowCrop", "fodder legumes (Fabaceae), for seed production (by contract),"),
        ZueriCropClass(70, "Unknown", "Vegetation", "Grassland", "Unknown", "fodder grasses for seed production (by contract),"),
        ZueriCropClass(71, "Meadow", "Vegetation", "Grassland", "Meadow", "Riverside meadows along rivers (without pastures),"),
        ZueriCropClass(72, "Unknown", "Vegetation", "Grassland", "BiodiversityArea", "Other green area (permanent green area),, entitled to contributions"),
        ZueriCropClass(73, "Unknown", "Vegetation", "Grassland", "Unknown", "Remaining green area (permanent green areas),, not eligible"),
        ZueriCropClass(74, "Vines", "Vegetation", "Orchards", "OrchardCrop", "vines"),
        ZueriCropClass(75, "Apples", "Vegetation", "Orchards", "OrchardCrop", "Fruit plants (apples),"),
        ZueriCropClass(76, "Pears", "Vegetation", "Orchards", "OrchardCrop", "Fruit plants (pears),"),
        ZueriCropClass(77, "StoneFruit", "Vegetation", "Orchards", "OrchardCrop", "Fruit plants (Steinobs),"),
        ZueriCropClass(78, "Berries", "Vegetation", "Special crops", "Berries", "Perennial berries"),
        ZueriCropClass(79, "Unknown", "Vegetation", "Special crops", "Unknown", "Perennial spice and medicinal plants"),
        ZueriCropClass(80, "Unknown", "Vegetation", "Special crops", "Unknown", "Perennial renewable resources (miscanthus, etc.),"),
        ZueriCropClass(81, "Hops", "Vegetation", "Orchards", "OrchardCrop", "hop"),
        ZueriCropClass(82, "Unknown", "Vegetation", "Special crops", "Unknown", "rhubarb"),
        ZueriCropClass(83, "Unknown", "Vegetation", "Special crops", "Unknown", "asparagus"),
        ZueriCropClass(84, "TreeCrop", "Vegetation", "Orchards", "TreeCrop", "Christmas trees"),
        ZueriCropClass(85, "TreeCrop", "Vegetation", "Orchards", "TreeCrop", "Nursery of forest plants outside the forest zone"),
        ZueriCropClass(86, "Unknown", "Vegetation", "Special crops", "Unknown", "Ornamental shrubs, ornamental shrubs and ornamental shrubs"),
        ZueriCropClass(87, "Unknown", "Vegetation", "Special crops", "Unknown", "Other nurseries (roses, fruits, etc.),"),
        ZueriCropClass(88, "Vines", "Vegetation", "Orchards", "OrchardCrop", "Vineyards with natural biodiversity"),
        ZueriCropClass(89, "Unknown", "Vegetation", "Special crops", "Unknown", "Truffle plants (in production),"),
        ZueriCropClass(90, "Unknown", "Vegetation", "Special crops", "Unknown", "Mulberry trees (feeding silkworms),"),
        ZueriCropClass(91, "Chestnut", "Vegetation", "Orchards", "OrchardCrop", "Cultivated selven (chestnut trees),"),
        ZueriCropClass(92, "Unknown", "Vegetation", "Special crops", "Unknown", "Perennial horticultural outdoor crops (not in the greenhouse),"),
        ZueriCropClass(93, "Vines", "Vegetation", "Orchards", "OrchardCrop", "vines nursery"),
        ZueriCropClass(94, "Unknown", "Vegetation", "Special crops", "Unknown", "Other fruit plants (kiwis, elderberries, etc.),"),
        ZueriCropClass(95, "Vines", "Vegetation", "Orchards", "OrchardCrop", "vines (region-specific biodiversity area),"),
        ZueriCropClass(96, "Unknown", "Vegetation", "Special crops", "BiodiversityArea", "Other areas with permanent crops, eligible"),
        ZueriCropClass(97, "Unknown", "Vegetation", "Special crops", "Unknown", "Other areas with permanent crops, not eligible"),
        ZueriCropClass(98, "Vegetables", "Infrastructure", "Unknown", "Greenhouse", "Vegetable crops in greenhouses with solid foundations"),
        ZueriCropClass(99, "Special cultures", "Infrastructure", "Unknown", "Greenhouse", "Other specialized crops in greenhouses with solid foundations"),
        ZueriCropClass(100, "Special cultures", "Infrastructure", "Unknown", "Greenhouse", "Horticultural crops in greenhouses with solid foundations"),
        ZueriCropClass(101, "Vegetables", "Unknown", "Unknown", "ProtectedCultivation", "Vegetable crops in protected cultivation without firm foundations"),
        ZueriCropClass(102, "Special cultures", "Unknown", "Unknown", "ProtectedCultivation", "Other special crops in protected cultivation without firm foundations"),
        ZueriCropClass(103, "Special cultures", "Unknown", "Unknown", "ProtectedCultivation", "Horticultural crops in protected cultivation without firm foundations"),
        ZueriCropClass(104, "Special cultures", "Unknown", "Unknown", "ProtectedCultivation", "Other crops in protected cultivation without a firm foundation"),
        ZueriCropClass(105, "Unknown", "Infrastructure", "Unknown", "Unknown", "other cultures in protected cultivation with solid foundation),"),
        ZueriCropClass(106, "Special cultures", "Unknown", "Unknown", "ProtectedCultivation", "Other crops in protected cultivation without firm foundations, not eligible"),
        ZueriCropClass(107, "Unknown", "Vegetation", "Special crops", "Unknown", "Scattering areas in the LN"),
        ZueriCropClass(108, "Hedge", "Vegetation", "Special crops", "Hedge", "Hedge, field and bank shrubs (with herbaceous area),"),
        ZueriCropClass(109, "Hedge", "Vegetation", "Special crops", "Hedge", "Hedgerow, field and bank shrubs (with buffer strips),"),
        ZueriCropClass(110, "Hedge", "Vegetation", "Special crops", "Hedge", "Hedgerow, field and bank shrubs (with buffer strips), (region-specific biodiversity production area),"),
        ZueriCropClass(111, "Multiple", "Undefined", "Unknown", "Unknown", "Other areas within the LN, entitled to contribute"),
        ZueriCropClass(112, "Multiple", "Undefined", "Unknown", "Unknown", "Other areas within the LN, not eligible"),
        ZueriCropClass(113, "Forest", "Vegetation", "Forest", "Forest", "Forest"),
        ZueriCropClass(114, "Multiple", "Vegetation", "Special crops", "Multiple", "Other unproductive areas (eg mulched areas, heavily weedy areas, hedges without buffer strips),"),
        ZueriCropClass(115, "Non agriculture", "Undefined", "Unknown", "Undefined", "Areas without main agricultural purpose (developed building land, playground, riding, camping, golf, air and military spaces"),
        ZueriCropClass(116, "Waters", "Undefined", "Unknown", "Unknown", "Ditches, ponds, ponds"),
        ZueriCropClass(117, "Non agriculture", "Infrastructure", "Unknown", "Unknown", "Ruderal areas, cairns and ramparts"),
        ZueriCropClass(118, "Multiple", "Infrastructure", "Unknown", "Unknown", "dry stone walls"),
        ZueriCropClass(119, "Unknown", "Bare soil", "Unknown", "Unknown", "non-asphalted, natural paths"),
        ZueriCropClass(120, "Biodiversity encouragement area", "Vegetation", "Grassland", "BiodiversityArea", "Region-specific biodiversity promotion areas"),
        ZueriCropClass(121, "Gardens", "Vegetation", "Special crops", "Gardens", "home gardens"),
        ZueriCropClass(122, "Unknown", "Infrastructure", "Unknown", "Unknown", "agricultural production in buildings (e.g. champignons, brussel sprouts),"),
        ZueriCropClass(123, "Pasture", "Vegetation", "Grassland", "Pasture", "Summer pastures"),
        ZueriCropClass(124, "Non agriculture", "Undefined", "Unknown", "Undefined", "Other areas outside the LN and SF")
    ]

    def __init__(
        self,
        root: str = ".data/zuericrop",
        transform: Compose = Compose([ToTensor()]),
    ):
        self.transform = transform
        self.f = h5py.File(os.path.join(root, "ZueriCrop.hdf5"), "r")

    def __len__(self) -> int:
        return self.f["data"].shape[0]

    def __getitem__(self, idx: int) -> Dict:
        x = self.f["data"][idx, ...]
        mask = self.f["gt"][idx, ...]
        instance_mask = self.f["gt_instance"][idx, ...]
        x, mask, instance_mask = self.transform([x, mask, instance_mask])
        return dict(x=x, mask=mask, instance_mask=instance_mask)



from typing import Optional

from torchrs.transforms import Compose, ToTensor
from torchrs.datasets.utils import dataset_split
from torchrs.train.datamodules import BaseDataModule
from torchrs.datasets import S2MTCP


class S2MTCPDataModule(BaseDataModule):

    def __init__(
        self,
        root: str = ".data/s2mtcp",
        transform: Compose = Compose([ToTensor()]),
        *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.root = root
        self.transform = transform

    def setup(self, stage: Optional[str] = None):
        dataset = S2MTCP(root=self.root, transform=self.transform)
        self.train_dataset, self.val_dataset, self.test_dataset = dataset_split(
            dataset, val_pct=self.val_split, test_pct=self.test_split
        )



from typing import Optional

from torchrs.train.datamodules import BaseDataModule
from torchrs.datasets import RSVQALR, RSVQAHR, RSVQAxBEN
from torchrs.transforms import Compose, ToTensor


class RSVQALRDataModule(BaseDataModule):

    def __init__(
        self,
        root: str = ".data/RSVQA_LR",
        image_transform: Compose = Compose([ToTensor()]),
        text_transform: Compose = Compose([]),
        *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.root = root
        self.image_transform = image_transform
        self.text_transform = text_transform

    def setup(self, stage: Optional[str] = None):
        self.train_dataset = RSVQALR(root=self.root, split="train", image_transform=self.image_transform, text_transform=self.text_transform)
        self.val_dataset = RSVQALR(root=self.root, split="val", image_transform=self.image_transform, text_transform=self.text_transform)
        self.test_dataset = RSVQALR(root=self.root, split="test", image_transform=self.image_transform, text_transform=self.text_transform)


class RSVQAHRDataModule(BaseDataModule):

    def __init__(
        self,
        root: str = ".data/RSVQA_HR",
        image_transform: Compose = Compose([ToTensor()]),
        text_transform: Compose = Compose([]),
        *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.root = root
        self.image_transform = image_transform
        self.text_transform = text_transform

    def setup(self, stage: Optional[str] = None):
        self.train_dataset = RSVQAHR(root=self.root, split="train", image_transform=self.image_transform, text_transform=self.text_transform)
        self.val_dataset = RSVQAHR(root=self.root, split="val", image_transform=self.image_transform, text_transform=self.text_transform)
        self.test_dataset = RSVQAHR(root=self.root, split="test", image_transform=self.image_transform, text_transform=self.text_transform)


class RSVQAxBENDataModule(BaseDataModule):

    def __init__(
        self,
        root: str = ".data/rsvqaxben",
        image_transform: Compose = Compose([ToTensor()]),
        text_transform: Compose = Compose([]),
        *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.root = root
        self.image_transform = image_transform
        self.text_transform = text_transform

    def setup(self, stage: Optional[str] = None):
        self.train_dataset = RSVQAxBEN(root=self.root, split="train", image_transform=self.image_transform, text_transform=self.text_transform)
        self.val_dataset = RSVQAxBEN(root=self.root, split="val", image_transform=self.image_transform, text_transform=self.text_transform)
        self.test_dataset = RSVQAxBEN(root=self.root, split="test", image_transform=self.image_transform, text_transform=self.text_transform)



import inspect
import pytest
from torch.utils.data import Dataset

from torchrs import datasets


skip = ["RSVQAxBEN", "S2MTCP"]


@pytest.mark.parametrize("dataset", reversed(datasets.__all__))
def test_datamodules(dataset: Dataset):

    if dataset in skip:
        return

    dataclass = getattr(datasets, dataset)

    if "split" in inspect.getfullargspec(dataclass).args:
        for split in dataclass.splits:
            ds = dataclass(split=split)
            length = len(ds)
            sample = ds[0]
            sample = ds[-1]
    else:
        ds = dataclass()
        length = len(ds)
        sample = ds[0]
        sample = ds[-1]



from typing import Optional

import torchvision.transforms as T

from torchrs.train.datamodules import BaseDataModule
from torchrs.datasets import SydneyCaptions


class SydneyCaptionsDataModule(BaseDataModule):

    def __init__(
        self,
        root: str = ".data/sydney_captions",
        transform: T.Compose = T.Compose([T.ToTensor()]),
        *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.root = root
        self.transform = transform

    def setup(self, stage: Optional[str] = None):
        self.train_dataset = SydneyCaptions(root=self.root, split="train", transform=self.transform)
        self.val_dataset = SydneyCaptions(root=self.root, split="val", transform=self.transform)
        self.test_dataset = SydneyCaptions(root=self.root, split="test", transform=self.transform)



from typing import Union, Any, Sequence, Callable, Tuple, List

import torch
import einops
import numpy as np
import torchvision.transforms as T


__all__ = ["Compose", "ToTensor", "ToDtype", "ExtractChips", "MinMaxNormalize", "Clip"]


class Compose(T.Compose):
    """ Custom Compose which processes a list of inputs """
    def __init__(self, transforms: Sequence[Callable]):
        self.transforms = transforms

    def __call__(self, x: Union[Any, Sequence]):
        if isinstance(x, Sequence):
            for t in self.transforms:
                x = [t(i) for i in x]
        else:
            for t in self.transforms:
                x = t(x)
        return x


class ToTensor(object):
    """ Custom ToTensor op which doesn't perform min-max normalization """
    def __init__(self, permute_dims: bool = True):
        self.permute_dims = permute_dims

    def __call__(self, x: np.ndarray) -> torch.Tensor:

        if x.dtype == "uint16":
            x = x.astype("int32")

        if isinstance(x, np.ndarray):
            x = torch.from_numpy(x)

        if x.ndim == 2:
            if self.permute_dims:
                x = x[:, :, None]
            else:
                x = x[None, :, :]

        # Convert HWC->CHW
        if self.permute_dims:
            if x.ndim == 4:
                x = x.permute((0, 3, 1, 2)).contiguous()
            else:
                x = x.permute((2, 0, 1)).contiguous()

        return x


class ToDtype(object):
    """ Convert input tensor to specified dtype """
    def __init__(self, dtype: torch.dtype):
        self.dtype = dtype

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return x.to(self.dtype)


class ExtractChips(object):
    """ Convert an tensor or ndarray into patches """
    def __init__(self, shape: Tuple[int, int]):
        self.h, self.w = shape

    def __call__(self, x: Union[torch.Tensor, np.ndarray]) -> Union[torch.Tensor, np.ndarray]:
        return einops.rearrange(x, "c (h p1) (w p2) -> (h w) c p1 p2", p1=self.h, p2=self.w)


class MinMaxNormalize(object):
    """ Normalize channels to the range [0, 1] using min/max values """
    def __init__(self, min: List[float], max: List[float]):
        self.min = torch.tensor(min)[:, None, None]
        self.max = torch.tensor(max)[:, None, None]
        self.denominator = (self.max - self.min)

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return (x - self.min) / self.denominator


class Clip(object):
    """ Clip channels to the range [min, max] """
    def __init__(self, min: List[float], max: List[float]):
        self.min = torch.tensor(min)[:, None, None]
        self.max = torch.tensor(max)[:, None, None]

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        x = torch.where(x < self.min, self.min, x)
        x = torch.where(x > self.max, self.max, x)
        return x



from typing import Optional, List

from torch.utils.data import Dataset, random_split


def dataset_split(dataset: Dataset, val_pct: float, test_pct: Optional[float] = None) -> List[Dataset]:
    """ Split a torch Dataset into train/val/test sets """
    if test_pct is None:
        val_length = int(len(dataset) * val_pct)
        train_length = len(dataset) - val_length
        return random_split(dataset, [train_length, val_length])
    else:
        val_length = int(len(dataset) * val_pct)
        test_length = int(len(dataset) * test_pct)
        train_length = len(dataset) - (val_length + test_length)
        return random_split(dataset, [train_length, val_length, test_length])



import os
from glob import glob
from typing import List, Tuple, Dict

import torch
import numpy as np
from PIL import Image

from torchrs.transforms import Compose, ToTensor


class DubaiSegmentation(torch.utils.data.Dataset):
    """ Semantic segmentation dataset of Dubai imagery taken by MBRSC satellites
    https://humansintheloop.org/resources/datasets/semantic-segmentation-dataset/

    """
    classes = {
        "Unlabeled":            {"rgb": (155, 155, 155), "color": "#9B9B9B"},
        "Water":                {"rgb": (226, 169, 41),  "color": "#E2A929"},
        "Land (unpaved area)":  {"rgb": (132, 41, 246),  "color": "#8429F6"},
        "Road":                 {"rgb": (110, 193, 228), "color": "#6EC1E4"},
        "Building":             {"rgb": (60, 16, 152),   "color": "#3C1098"},
        "Vegetation":           {"rgb": (254, 221, 58),  "color": "#FEDD3A"}
    }
    colors = [v["rgb"] for k, v in classes.items()]

    def __init__(
        self,
        root: str = ".data/dubai-segmentation",
        transform: Compose = Compose([ToTensor()]),
    ):
        self.transform = transform
        self.images = self.load_images(root)
        self.regions = list(set([image["region"] for image in self.images]))

    @staticmethod
    def load_images(path: str) -> List[Dict]:
        images = sorted(glob(os.path.join(path, "**", "images", "*.jpg"), recursive=True))
        masks = sorted(glob(os.path.join(path, "**", "masks", "*.png"), recursive=True))
        regions = [image.split(os.sep)[-3] for image in images]
        files = [
            dict(image=image, mask=mask, region=region)
            for image, mask, region in zip(images, masks, regions)
        ]
        return files

    @staticmethod
    def rgb_to_mask(rgb: np.ndarray, colors: List[Tuple[int, int, int]]) -> np.ndarray:
        h, w = rgb.shape[:2]
        mask = np.zeros(shape=(h, w), dtype=np.uint8)
        for i, c in enumerate(colors):
            cmask = (rgb == c)
            if isinstance(cmask, np.ndarray):
                mask[cmask.all(axis=-1)] = i

        return mask

    def __len__(self) -> int:
        return len(self.images)

    def __getitem__(self, idx: int) -> Dict:
        image_path, target_path = self.images[idx]["image"], self.images[idx]["mask"]
        x = np.array(Image.open(image_path).convert("RGB"))
        y = np.array(Image.open(target_path).convert("RGB"))
        y = self.rgb_to_mask(y, self.colors)
        x, y = self.transform([x, y])
        return dict(x=x, mask=y, region=self.images[idx]["region"])



import os

import torchvision.transforms as T

from torchrs.datasets import RSICD


class UCMCaptions(RSICD):
    """ UC Merced (UCM) Captions dataset from 'Deep semantic understanding of
    high resolution remote sensing image', Qu et al (2016)
    https://ieeexplore.ieee.org/document/7546397

    'The UCM dataset totally has 2100 HSR images which are divided into 21 challenging
    scene categories ... Then every HSR image is annotated with 5 reference sentences
    ... one has totally 2100 HSR remote sensing images with 10500 descriptions'
    """
    splits = ["train", "val", "test"]

    def __init__(
        self,
        root: str = ".data/ucm_captions",
        split: str = "train",
        transform: T.Compose = T.Compose([T.ToTensor()])
    ):
        assert split in self.splits
        self.root = root
        self.transform = transform
        self.captions = self.load_captions(os.path.join(root, "dataset.json"), split)
        self.image_root = "images"



import os
from glob import glob
from typing import Tuple, List

import torch
import numpy as np
import torchvision.transforms as T
from PIL import Image


class BrazilianCoffeeScenes(torch.utils.data.Dataset):
    """ Brazilian Coffee Scenes dataset from 'Do Deep Features Generalize from
    Everyday Objects to Remote Sensing and Aerial Scenes Domains?', Penatti at al. (2015)
    https://arxiv.org/abs/1703.00121
    """
    bands = ["Green", "Red", "NIR"]
    classes = ["non-coffee", "coffee"]

    def __init__(
        self,
        root: str = ".data/brazilian_coffee_scenes",
        transform: T.Compose = T.Compose([T.ToTensor()])
    ):
        super().__init__()
        self.transform = transform
        self.images, self.labels = self.load_images(root)

    @staticmethod
    def load_images(path: str) -> Tuple[List[str], List[int]]:
        folds = glob(os.path.join(path, "*.txt"))
        images, labels = [], []
        for fold in folds:
            fold_dir = os.path.join(path, os.path.splitext(os.path.basename(fold))[0])
            with open(fold, "r") as f:
                lines = f.read().strip().splitlines()

            for line in lines:
                label, image = line.split(".", 1)
                images.append(os.path.join(fold_dir, image + ".jpg"))
                labels.append(0 if label == "noncoffee" else 1)

        return images, labels

    def __len__(self) -> int:
        return len(self.images)

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, int]:
        image, y = self.images[idx], self.labels[idx]
        x = np.array(Image.open(image).convert("RGB"))
        x = self.transform(x)
        return x, y



import os
from glob import glob
from typing import List, Dict

import torch
import numpy as np
from PIL import Image

from torchrs.transforms import Compose, ToTensor


class ETCI2021(torch.utils.data.Dataset):
    """Sentinel-1 Synthetic Aperature Radar (SAR) segmentation dataset from the ETCI 2021 Competition on Flood Detection
    https://nasa-impact.github.io/etci2021/
    https://competitions.codalab.org/competitions/30440

    'The contest dataset is composed of 66,810 (33,405 x 2 VV & VH polarization) tiles of 256 x 256 pixels,
    distributed respectively across the training, validation and test sets as follows: 33,405, 10,400,
    and 12,348 tiles for each polarization. Each tile includes 3 RGB channels which have been converted
    by tiling 54 labeled GeoTiff files generated from Sentinel-1 C-band synthetic aperture radar (SAR)
    imagery data using Hybrid Pluggable Processing Pipeline hyp3.'

    Note that hyp3 preprocessing generates 3 band for each band so VV and VH are both of shape (256, 256, 3)
    """
    bands = ["VV", "VH"]
    splits = ["train", "val", "test"]
    split_to_folder = dict(train="train", val="test", test="test_internal")

    def __init__(
        self,
        root: str = ".data/etci2021",
        split: str = "train",
        transform: Compose = Compose([ToTensor()]),
    ):
        assert split in self.splits
        self.split = split
        self.transform = transform
        self.images = self.load_files(root, self.split_to_folder[split])

    @staticmethod
    def load_files(root: str, split: str) -> List[Dict]:
        images = []
        folders = sorted(glob(os.path.join(root, split, "*")))
        folders = [f + "/tiles" for f in folders]
        for folder in folders:
            vvs = glob(os.path.join(folder, "vv", "*.png"))
            vhs = glob(os.path.join(folder, "vh", "*.png"))
            water_masks = glob(os.path.join(folder, "water_body_label", "*.png"))

            if split == "test_internal":
                flood_masks = [None] * len(water_masks)
            else:
                flood_masks = glob(os.path.join(folder, "flood_label", "*.png"))

            for vv, vh, flood_mask, water_mask in zip(vvs, vhs, flood_masks, water_masks):
                images.append(dict(vv=vv, vh=vh, flood_mask=flood_mask, water_mask=water_mask))
        return images

    def __len__(self) -> int:
        return len(self.images)

    def __getitem__(self, idx: int) -> Dict[str, torch.Tensor]:
        """ Returns a dict containing vv, vh, flood mask, water mask
        vv: (3, h, w)
        vh: (3, h, w)
        flood mask: (1, h, w) flood mask
        water mask: (1, h, w) water mask
        """
        images = self.images[idx]
        vv = np.array(Image.open(images["vv"]), dtype="uint8")
        vh = np.array(Image.open(images["vh"]), dtype="uint8")
        water_mask = np.array(Image.open(images["water_mask"]).convert("L"), dtype="bool")

        if self.split == "test":
            vv, vh, water_mask = self.transform([vv, vh, water_mask])
            output = dict(vv=vv, vh=vh, water_mask=water_mask)
        else:
            flood_mask = np.array(Image.open(images["flood_mask"]).convert("L"), dtype="bool")
            vv, vh, flood_mask, water_mask = self.transform([vv, vh, flood_mask, water_mask])
            output = dict(vv=vv, vh=vh, flood_mask=flood_mask, water_mask=water_mask)

        return output



import torch
import torch.nn as nn
from einops import rearrange


class ConvBlock(nn.Sequential):

    def __init__(self, in_channels: int, out_channels: int, kernel_size: int = 3, dropout: float = 0.2, norm: bool = True):
        super().__init__(
            nn.Conv2d(in_channels, out_channels, kernel_size),
            nn.BatchNorm2d(out_channels) if norm else nn.Identity(),
            nn.ReLU(),
            nn.Dropout(dropout)
        )


class EarlyFusion(nn.Module):
    """ Early Fusion (EF) from 'Urban Change Detection for Multispectral Earth Observation
    Using ConvolutionalNeural Networks', Daudt et al. (2018)
    https://arxiv.org/abs/1810.08468

    This model takes as input the concatenated image pairs (T*C, 15, 15)
    and is essentially a simple CNN classifier of the central pixel in an input patch.
    Assumes (T*Cx15x15) patch size input
    """
    def __init__(self, channels: int = 3, t: int = 2, num_classes: int = 2):
        super().__init__()
        filters = [channels * t, 32, 32, 64, 64, 128, 128]
        dropout = 0.2
        self.encoder = nn.Sequential(
            *[ConvBlock(filters[i-1], filters[i]) for i in range(1, len(filters))],
            ConvBlock(filters[-1], 128, dropout=0.0, norm=False),
            nn.Flatten()
        )
        self.mlp = nn.Sequential(
            nn.Linear(128, 8),
            nn.BatchNorm1d(8),
            nn.ReLU(),
            nn.Dropout2d(dropout),
            nn.Linear(8, num_classes)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = rearrange(x, "b t c h w -> b (t c) h w")
        x = self.encoder(x)
        x = self.mlp(x)
        return x


class Siam(nn.Module):
    """ Siamese (Siam) from 'Urban Change Detection for Multispectral Earth Observation
    Using ConvolutionalNeural Networks', Daudt et al. (2018)
    https://arxiv.org/abs/1810.08468

    This model takes as input the concatenated image pairs (T*C, 15, 15)
    and is essentially a simple CNN classifier of the central pixel in an input patch.
    Assumes (T*Cx15x15) patch size input
    """
    def __init__(self, channels: int = 3, t: int = 2, num_classes: int = 2):
        super().__init__()
        filters = [channels, 64, 64, 128]
        dropout = 0.2
        self.encoder = nn.Sequential(
            *[ConvBlock(filters[i-1], filters[i]) for i in range(1, len(filters))],
            ConvBlock(filters[-1], 128, dropout=0.0),
        )
        self.mlp = nn.Sequential(
            nn.Linear(t*128*7*7, 64),
            nn.BatchNorm1d(64),
            nn.ReLU(),
            nn.Dropout2d(dropout),
            nn.Linear(64, num_classes)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        b, t, c, h, w = x.shape
        x = rearrange(x, "b t c h w -> (b t) c h w")
        x = self.encoder(x)
        x = rearrange(x, "(b t) c h w -> b (t c h w)", b=b)
        x = self.mlp(x)
        return x



import os
from glob import glob
from typing import Dict

import torch
import numpy as np
from PIL import Image

from torchrs.transforms import Compose, ToTensor


class S2Looking(torch.utils.data.Dataset):
    """ The Satellite Side-Looking (S2Looking) dataset from 'S2Looking: A Satellite Side-Looking
    Dataset for Building Change Detection', Shen at al. (2021)
    https://arxiv.org/abs/2107.09244

    'S2Looking is a building change detection dataset that contains large-scale side-looking
    satellite images captured at varying off-nadir angles. The S2Looking dataset consists of
    5,000 registered bitemporal image pairs (size of 1024*1024, 0.5 ~ 0.8 m/pixel) of rural
    areas throughout the world and more than 65,920 annotated change instances. We provide
    two label maps to separately indicate the newly built and demolished building regions
    for each sample in the dataset.'
    """
    splits = ["train", "val", "test"]

    def __init__(
        self,
        root: str = ".data/s2looking",
        split: str = "train",
        transform: Compose = Compose([ToTensor()]),
    ):
        assert split in self.splits
        self.root = root
        self.transform = transform
        self.files = self.load_files(root, split)

    @staticmethod
    def load_files(root: str, split: str):
        files = []
        images = glob(os.path.join(root, split, "Image1", "*.png"))
        images = sorted([os.path.basename(image) for image in images])
        for image in images:
            image1 = os.path.join(root, split, "Image1", image)
            image2 = os.path.join(root, split, "Image2", image)
            build_mask = os.path.join(root, split, "label1", image)
            demo_mask = os.path.join(root, split, "label2", image)
            files.append(dict(image1=image1, image2=image2, build_mask=build_mask, demolish_mask=demo_mask))
        return files

    def __len__(self) -> int:
        return len(self.files)

    def __getitem__(self, idx: int) -> Dict:
        """ Returns a dict containing x, mask
        x: (2, 13, h, w)
        build_mask: (1, h, w)
        demolish_mask: (1, h, w)
        """
        files = self.files[idx]
        build_mask = np.array(Image.open(files["build_mask"]))
        demo_mask = np.array(Image.open(files["demolish_mask"]))
        build_mask = np.clip(build_mask.mean(axis=-1), 0, 1).astype("uint8")
        demo_mask = np.clip(demo_mask.mean(axis=-1), 0, 1).astype("uint8")
        image1 = np.array(Image.open(files["image1"]))
        image2 = np.array(Image.open(files["image2"]))
        image1, image2, build_mask, demo_mask = self.transform([image1, image2, build_mask, demo_mask])
        x = torch.stack([image1, image2], dim=0)
        return dict(x=x, build_mask=build_mask, demolish_mask=demo_mask)



from . import utils
from .probav import PROBAV
from .etci2021 import ETCI2021
from .rsvqa import RSVQALR, RSVQAHR, RSVQAxBEN
from .eurosat import EuroSATRGB, EuroSATMS
from .resisc45 import RESISC45
from .rsicd import RSICD
from .oscd import OSCD
from .s2looking import S2Looking
from .levircd import LEVIRCDPlus
from .fair1m import FAIR1M
from .sydney_captions import SydneyCaptions
from .ucm_captions import UCMCaptions
from .s2mtcp import S2MTCP
from .advance import ADVANCE
from .sat import SAT4, SAT6
from .hrscd import HRSCD
from .inria_ail import InriaAIL
from .tiselac import Tiselac
from .gid15 import GID15
from .zuericrop import ZueriCrop
from .aid import AID
from .dubai_segmentation import DubaiSegmentation
from .hkh_glacier import HKHGlacierMapping
from .ucm import UCM
from .patternnet import PatternNet
from .whu_rs19 import WHURS19
from .rsscn7 import RSSCN7
from .brazilian_coffee import BrazilianCoffeeScenes


__all__ = [
    "PROBAV", "ETCI2021", "RSVQALR", "RSVQAxBEN", "EuroSATRGB", "EuroSATMS",
    "RESISC45", "RSICD", "OSCD", "S2Looking", "LEVIRCDPlus", "FAIR1M",
    "SydneyCaptions", "UCMCaptions", "S2MTCP", "ADVANCE", "SAT4", "SAT6",
    "HRSCD", "InriaAIL", "Tiselac", "GID15", "ZueriCrop", "AID", "DubaiSegmentation",
    "HKHGlacierMapping", "UCM", "PatternNet", "RSVQAHR", "WHURS19", "RSSCN7",
    "BrazilianCoffeeScenes"
]



from typing import Dict

import torch
import torch.nn as nn
import torchmetrics
import pytorch_lightning as pl

from torchrs.models import FCEF, FCSiamDiff, FCSiamConc


class BaseFCCDModule(pl.LightningModule):

    def __init__(
        self,
        channels: int = 3,
        t: int = 2,
        num_classes: int = 2,
        loss_fn: nn.Module = nn.CrossEntropyLoss(),
        opt: torch.optim.Optimizer = torch.optim.Adam,
        lr: float = 3E-4
    ):
        super().__init__()
        self.loss_fn = loss_fn
        self.opt = opt
        self.lr = lr

        metrics = torchmetrics.MetricCollection([
            torchmetrics.Accuracy(threshold=0.5, num_classes=num_classes, average="micro", mdmc_average="global"),
            torchmetrics.Precision(num_classes=num_classes, threshold=0.5, average="micro", mdmc_average="global"),
            torchmetrics.Recall(num_classes=num_classes, threshold=0.5, average="micro", mdmc_average="global"),
            torchmetrics.F1(num_classes=num_classes, threshold=0.5, average="micro", mdmc_average="global"),
            torchmetrics.IoU(threshold=0.5, num_classes=num_classes),
        ])
        self.train_metrics = metrics.clone(prefix='train_')
        self.val_metrics = metrics.clone(prefix='val_')
        self.test_metrics = metrics.clone(prefix='test_')

    def configure_optimizers(self):
        return self.opt(self.parameters(), lr=self.lr)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)

    def training_step(self, batch: Dict, batch_idx: int):
        x, y = batch
        y_pred = self(x)
        loss = self.loss_fn(y_pred, y)
        metrics = self.train_metrics(y_pred.softmax(dim=1), y)
        metrics["train_loss"] = loss
        self.log_dict(metrics)
        return loss

    def validation_step(self, batch: Dict, batch_idx: int):
        x, y = batch
        y_pred = self(x)
        loss = self.loss_fn(y_pred, y)
        metrics = self.val_metrics(y_pred.softmax(dim=1), y)
        metrics["val_loss"] = loss
        self.log_dict(metrics)

    def test_step(self, batch: Dict, batch_idx: int):
        x, y = batch
        y_pred = self(x)
        loss = self.loss_fn(y_pred, y)
        metrics = self.test_metrics(y_pred.softmax(dim=1), y)
        metrics["test_loss"] = loss
        self.log_dict(metrics)


class FCEFModule(BaseFCCDModule):

    def __init__(
        self,
        channels: int = 3,
        t: int = 2,
        num_classes: int = 2,
        *args, **kwargs
    ):
        super().__init__(channels, t, num_classes, *args, **kwargs)
        self.model = FCEF(channels, t, num_classes)


class FCSiamConcModule(BaseFCCDModule):

    def __init__(
        self,
        channels: int = 3,
        t: int = 2,
        num_classes: int = 2,
        *args, **kwargs
    ):
        super().__init__(channels, t, num_classes, *args, **kwargs)
        self.model = FCSiamConc(channels, t, num_classes)


class FCSiamDiffModule(BaseFCCDModule):

    def __init__(
        self,
        channels: int = 3,
        t: int = 2,
        num_classes: int = 2,
        *args, **kwargs
    ):
        super().__init__(channels, t, num_classes, *args, **kwargs)
        self.model = FCSiamDiff(channels, t, num_classes)



import itertools
import pytest
import torch

from torchrs.models import RAMS


DTYPE = torch.float32
DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
IMAGE_SIZE = 32
T = [9, 11, 13, 15]
SCALE_FACTOR = [2, 3, 4]
CHANNELS = [1, 3]
BATCH_SIZE = [1, 2]

params = list(itertools.product(SCALE_FACTOR, T, CHANNELS, BATCH_SIZE))


@torch.no_grad()
@pytest.mark.parametrize("scale_factor, t, channels, batch_size", params)
def test_rams(scale_factor, t, channels, batch_size):
    model = RAMS(scale_factor, t, channels, num_feature_attn_blocks=3)
    model = model.to(DEVICE)
    model = model.eval()
    lr = torch.ones(batch_size, t, channels, IMAGE_SIZE, IMAGE_SIZE)
    lr = lr.to(DTYPE)
    lr = lr.to(DEVICE)
    sr = model(lr)
    assert sr.shape == (batch_size, channels, IMAGE_SIZE*scale_factor, IMAGE_SIZE*scale_factor)
    assert sr.dtype == torch.float32



from typing import List, Tuple

import torch
import torch.nn as nn
from einops import rearrange


class ConvBlock(nn.Module):

    def __init__(self, filters: List[int], kernel_size: int = 3, dropout: float = 0.2, pool: bool = True):
        super().__init__()
        layers = []
        for i in range(1, len(filters)):
            layers.extend([
                nn.Conv2d(filters[i - 1], filters[i], kernel_size, stride=1, padding=kernel_size//2),
                nn.BatchNorm2d(filters[i]),
                nn.ReLU(),
                nn.Dropout(dropout),
            ])
        self.model = nn.Sequential(*layers)
        self.pool = nn.MaxPool2d(kernel_size=2) if pool else nn.Identity()

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        x = self.model(x)
        return self.pool(x), x


class DeConvBlock(nn.Sequential):

    def __init__(self, filters: List[int], kernel_size: int = 3, dropout: float = 0.2):
        super().__init__(
            *[nn.Sequential(
                nn.ConvTranspose2d(filters[i - 1], filters[i], kernel_size, padding=kernel_size//2),
                nn.BatchNorm2d(filters[i]),
                nn.ReLU(),
                nn.Dropout(dropout)
            ) for i in range(1, len(filters))]
        )


class UpsampleBlock(nn.Sequential):

    def __init__(self, channels: int, kernel_size: int = 3):
        super().__init__(
            nn.ConvTranspose2d(channels, channels, kernel_size, padding=kernel_size//2, stride=2, output_padding=1)
        )


class Encoder(nn.ModuleList):

    def __init__(self, in_channels: int = 3):
        super().__init__([
            ConvBlock([in_channels, 16, 16]),
            ConvBlock([16, 32, 32]),
            ConvBlock([32, 64, 64, 64]),
            ConvBlock([64, 128, 128, 128])
        ])


class Decoder(nn.ModuleList):

    def __init__(self, num_classes: int = 2):
        super().__init__([
            DeConvBlock([256, 128, 128, 64]),
            DeConvBlock([128, 64, 64, 32]),
            DeConvBlock([64, 32, 16]),
            DeConvBlock([32, 16, num_classes])
        ])


class SiamEncoder(nn.ModuleList):

    def __init__(self, in_channels: int = 3):
        super().__init__([
            ConvBlock([in_channels, 16, 16]),
            ConvBlock([16, 32, 32]),
            ConvBlock([32, 64, 64, 64]),
            ConvBlock([64, 128, 128, 128], pool=False)
        ])


class ConcatDecoder(nn.ModuleList):

    def __init__(self, t: int = 2, num_classes: int = 2):
        scale = 0.5 * (t + 1)
        super().__init__([
            DeConvBlock([int(256 * scale), 128, 128, 64]),
            DeConvBlock([int(128 * scale), 64, 64, 32]),
            DeConvBlock([int(64 * scale), 32, 16]),
            DeConvBlock([int(32 * scale), 16, num_classes])
        ])


class Upsample(nn.ModuleList):

    def __init__(self):
        super().__init__([
            UpsampleBlock(128),
            UpsampleBlock(64),
            UpsampleBlock(32),
            UpsampleBlock(16)
        ])


class FCEF(nn.Module):
    """ Fully-convolutional Early Fusion (FC-EF) from
    'Fully Convolutional Siamese Networks for Change Detection', Daudt et al. (2018)
    https://arxiv.org/abs/1810.08462
    """
    def __init__(self, channels: int = 3, t: int = 2, num_classes: int = 2):
        super().__init__()
        self.encoder = Encoder(channels * t)
        self.decoder = Decoder(num_classes)
        self.upsample = Upsample()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        b, t, c, h, w = x.shape
        x = rearrange(x, "b t c h w -> b (t c) h w")

        skips = []
        for block in self.encoder:
            x, skip = block(x)
            skips.append(skip)

        for block, upsample, skip in zip(self.decoder, self.upsample, reversed(skips)):
            x = upsample(x)
            x = rearrange([x, skip], "t b c h w -> b (t c) h w")
            x = block(x)

        return x


class FCSiamConc(nn.Module):
    """ Fully-convolutional Siamese Concatenation (FC-Siam-conc) from
    'Fully Convolutional Siamese Networks for Change Detection', Daudt et al. (2018)
    https://arxiv.org/abs/1810.08462
    """
    def __init__(self, channels: int = 3, t: int = 2, num_classes: int = 2):
        super().__init__()
        self.encoder = SiamEncoder(channels)
        self.decoder = ConcatDecoder(t, num_classes)
        self.upsample = Upsample()
        self.pool = nn.MaxPool2d(kernel_size=2)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        b, t, c, h, w = x.shape
        x = rearrange(x, "b t c h w -> (b t) c h w")

        skips = []
        for block in self.encoder:
            x, skip = block(x)
            skips.append(skip)

        # Concat skips
        skips = [rearrange(skip, "(b t) c h w -> b (t c) h w", t=t) for skip in skips]

        # Only first input encoding is passed directly to decoder
        x = rearrange(x, "(b t) c h w -> b t c h w", t=t)
        x = x[:, 0, ...]
        x = self.pool(x)

        for block, upsample, skip in zip(self.decoder, self.upsample, reversed(skips)):
            x = upsample(x)
            x = torch.cat([x, skip], dim=1)
            x = block(x)

        return x


class FCSiamDiff(nn.Module):
    """ Fully-convolutional Siamese Difference (FC-Siam-diff) from
    'Fully Convolutional Siamese Networks for Change Detection', Daudt et al. (2018)
    https://arxiv.org/abs/1810.08462
    """
    def __init__(self, channels: int = 3, t: int = 2, num_classes: int = 2):
        super().__init__()
        self.encoder = SiamEncoder(channels)
        self.decoder = Decoder(num_classes)
        self.upsample = Upsample()
        self.pool = nn.MaxPool2d(kernel_size=2)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        b, t, c, h, w = x.shape
        x = rearrange(x, "b t c h w -> (b t) c h w")

        skips = []
        for block in self.encoder:
            x, skip = block(x)
            skips.append(skip)

        # Diff skips
        skips = [rearrange(skip, "(b t) c h w -> b t c h w", t=t) for skip in skips]
        diffs = []
        for skip in skips:
            diff, xt = skip[:, 0, ...], skip[:, 1:, ...]
            for i in range(t - 1):
                diff = torch.abs(diff - xt[:, i, ...])
                diffs.append(diff)

        # Only first input encoding is passed directly to decoder
        x = rearrange(x, "(b t) c h w -> b t c h w", t=t)
        x = x[:, 0, ...]
        x = self.pool(x)

        for block, upsample, skip in zip(self.decoder, self.upsample, reversed(diffs)):
            x = upsample(x)
            x = torch.cat([x, skip], dim=1)
            x = block(x)

        return x



import os
from glob import glob
from typing import Dict

import torch
import numpy as np
from PIL import Image

from torchrs.transforms import Compose, ToTensor


class LEVIRCDPlus(torch.utils.data.Dataset):
    """ LEVIR-CD+ dataset from 'S2Looking: A Satellite Side-Looking
    Dataset for Building Change Detection', Shen at al. (2021)
    https://arxiv.org/abs/2107.09244

    'LEVIR-CD+ contains more than 985 VHR (0.5m/pixel) bitemporal Google
    Earth images with dimensions of 1024x1024 pixels. These bitemporal images
    are from 20 different regions located in several cities in the state of
    Texas in the USA. The capture times of the image data vary from 2002 to
    2020. Images of different regions were taken at different times. The
    bitemporal images have a time span of 5 years.'
    """
    splits = ["train", "test"]

    def __init__(
        self,
        root: str = ".data/levircd_plus",
        split: str = "train",
        transform: Compose = Compose([ToTensor()]),
    ):
        assert split in self.splits
        self.root = root
        self.transform = transform
        self.files = self.load_files(root, split)

    @staticmethod
    def load_files(root: str, split: str):
        files = []
        images = glob(os.path.join(root, split, "A", "*.png"))
        images = sorted([os.path.basename(image) for image in images])
        for image in images:
            image1 = os.path.join(root, split, "A", image)
            image2 = os.path.join(root, split, "B", image)
            mask = os.path.join(root, split, "label", image)
            files.append(dict(image1=image1, image2=image2, mask=mask))
        return files

    def __len__(self) -> int:
        return len(self.files)

    def __getitem__(self, idx: int) -> Dict:
        """ Returns a dict containing x, mask
        x: (2, 13, h, w)
        mask: (1, h, w)
        """
        files = self.files[idx]
        mask = np.array(Image.open(files["mask"]))
        mask = np.clip(mask, 0, 1)
        image1 = np.array(Image.open(files["image1"]))
        image2 = np.array(Image.open(files["image2"]))
        image1, image2, mask = self.transform([image1, image2, mask])
        x = torch.stack([image1, image2], dim=0)
        return dict(x=x, mask=mask)



import os

import torchvision.transforms as T
from torchvision.datasets import ImageFolder


class PatternNet(ImageFolder):
    """ PatternNet dataset from 'PatternNet: A benchmark dataset for performance
    evaluation of remote sensing image retrieval', Zhou at al. (2018)
    https://arxiv.org/abs/1706.03424

    """
    def __init__(
        self,
        root: str = ".data/PatternNet",
        transform: T.Compose = T.Compose([T.ToTensor()])
    ):
        super().__init__(
            root=os.path.join(root, "images"),
            transform=transform
        )



import os

import torchvision.transforms as T
from torchvision.datasets import ImageFolder


class UCM(ImageFolder):
    """ UC Merced Land Use dataset from 'Bag-Of-Visual-Words and
    Spatial Extensions for Land-Use Classification', Yang at al. (2010)
    https://faculty.ucmerced.edu/snewsam/papers/Yang_ACMGIS10_BagOfVisualWords.pdf

    """
    def __init__(
        self,
        root: str = ".data/UCMerced_LandUse",
        transform: T.Compose = T.Compose([T.ToTensor()])
    ):
        super().__init__(
            root=os.path.join(root, "Images"),
            transform=transform
        )



import pytest
import torch

from torchrs.transforms import ExtractChips


def test_extract_chips():
    x = torch.ones(3, 128, 128)
    f = ExtractChips((32, 32))
    z = f(x)
    assert z.shape == (16, 3, 32, 32)



""" Referenced from official TF implementation https://github.com/EscVM/RAMS/blob/master/utils/network.py """
import torch
import torch.nn as nn
from einops import rearrange
from einops.layers.torch import Rearrange, Reduce


class ReflectionPad3d(nn.Module):
    """ Custom 3D reflection padding for only h, w dims """
    def __init__(self, padding):
        super().__init__()
        self.pad = nn.ReflectionPad2d(padding)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        b, t, c, h, w = x.shape
        x = rearrange(x, "b t c h w -> b (t c) h w")
        x = self.pad(x)
        x = rearrange(x, "b (t c) h w -> b t c h w", t=t, c=c)
        return x


class TemporalAttention(nn.Module):
    """ Temporal Attention Block """
    def __init__(self, channels: int, kernel_size: int, r: int):
        super().__init__()
        self.model = nn.Sequential(
            Reduce("b c h w -> b c () ()", "mean"),
            nn.Conv2d(channels, channels//r, kernel_size, stride=1, padding=kernel_size//2),
            nn.ReLU(),
            nn.Conv2d(channels//r, channels, kernel_size, stride=1, padding=kernel_size//2),
            nn.Sigmoid()
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * self.model(x)


class FeatureAttention(nn.Module):
    """ Feature Attention Block """
    def __init__(self, channels: int, kernel_size: int, r: int):
        super().__init__()
        self.model = nn.Sequential(
            Reduce("b c t h w -> b c () () ()", "mean"),
            nn.Conv3d(channels, channels//r, kernel_size, stride=1, padding=kernel_size//2),
            nn.ReLU(),
            nn.Conv3d(channels//r, channels, kernel_size, stride=1, padding=kernel_size//2),
            nn.Sigmoid()
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * self.model(x)


class RTAB(nn.Module):
    """ Residual Temporal Attention Block """
    def __init__(self, channels: int, kernel_size: int, r: int):
        super().__init__()
        self.model = nn.Sequential(
            nn.Conv2d(channels, channels, kernel_size, stride=1, padding=kernel_size // 2),
            nn.ReLU(),
            nn.Conv2d(channels, channels, kernel_size, stride=1, padding=kernel_size // 2),
            TemporalAttention(channels, kernel_size, r)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x + self.model(x)


class RFAB(nn.Module):
    """ Residual Feature Attention Block """
    def __init__(self, channels: int, kernel_size: int, r: int):
        super().__init__()
        self.model = nn.Sequential(
            nn.Conv3d(channels, channels, kernel_size, stride=1, padding=kernel_size // 2),
            nn.ReLU(),
            nn.Conv3d(channels, channels, kernel_size, stride=1, padding=kernel_size // 2),
            FeatureAttention(channels, kernel_size, r)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x + self.model(x)


class TemporalReductionBlock(nn.Module):
    """ Temporal Reduction Block """
    def __init__(self, channels: int, kernel_size: int, r: int):
        super().__init__()
        self.model = nn.Sequential(
            ReflectionPad3d(1),
            RFAB(channels, kernel_size, r),
            nn.Conv3d(channels, channels, kernel_size, stride=1, padding=0),
            nn.ReLU()
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)


class RAMS(nn.Module):
    """
    Residual Attention Multi-image Super-resolution Network (RAMS)
    'Multi-Image Super Resolution of Remotely Sensed Images Using Residual Attention Deep Neural Networks'
    Salvetti et al. (2021)
    https://www.mdpi.com/2072-4292/12/14/2207

    Note this model was built to work with t=9 input images and kernel_size=3. Other values may not work.
    t must satisfy the constraints of ((t-1)/(kernel_size-1) - 1) % 1 == 0 where kernel_size=3 and t >= 5.
    Some valid t's are [9, 11, 13, ...]
    """
    def __init__(
        self,
        scale_factor: int = 3,
        t: int = 9,
        c: int = 1,
        num_feature_attn_blocks: int = 12,
    ):
        super().__init__()
        filters = 32
        kernel_size = 3
        r = 8
        num_temporal_redn_blocks = ((t-1)/(kernel_size-1) - 1)
        err = """t must satisfy the ((t-1)/(kernel_size-1) - 1) % 1 == 0 where kernel_size=3
                and t >= 5. Some valid t's are [9, 11, 13, 15, ...] """
        assert num_temporal_redn_blocks % 1 == 0 and t >= 9, err

        self.temporal_attn = nn.Sequential(
            Rearrange("b t c h w -> b (t c) h w"),
            nn.ReflectionPad2d(1),
            RTAB(t * c, kernel_size, r),
        )
        self.residual_upsample = nn.Sequential(
            nn.Conv2d(t * c, c * scale_factor ** 2, kernel_size, stride=1, padding=0),
            nn.PixelShuffle(scale_factor)
        )
        self.head = nn.Sequential(
            Rearrange("b t c h w -> b c t h w"),
            ReflectionPad3d(1),
            nn.Conv3d(c, filters, kernel_size, stride=1, padding=kernel_size//2)
        )
        self.feature_attn = nn.Sequential(
            *[RFAB(filters, kernel_size, r) for _ in range(num_feature_attn_blocks)],
            nn.Conv3d(filters, filters, kernel_size, stride=1, padding=kernel_size//2)
        )
        self.temporal_redn = nn.Sequential(
            *[TemporalReductionBlock(filters, kernel_size, r) for _ in range(int(num_temporal_redn_blocks))]
        )
        self.main_upsample = nn.Sequential(
            nn.Conv3d(filters, c * scale_factor ** 2, kernel_size, stride=1, padding=0),
            Rearrange("b c t h w -> b (c t) h w"),
            nn.PixelShuffle(scale_factor),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Main branch
        h = self.head(x)
        feature_attn = h + self.feature_attn(h)
        temporal_redn = self.temporal_redn(feature_attn)
        temporal_redn = self.main_upsample(temporal_redn)

        # Global residual branch
        temporal_attn = self.temporal_attn(x)
        temporal_attn = self.residual_upsample(temporal_attn)

        return temporal_attn + temporal_redn



import os
import json
from typing import List, Dict

import torch
import torchvision.transforms as T
from PIL import Image


class RSICD(torch.utils.data.Dataset):
    """ Image Captioning Dataset from 'Exploring Models and Data for
    Remote Sensing Image Caption Generation', Lu et al. (2017)
    https://arxiv.org/abs/1712.07835

    'RSICD is used for remote sensing image captioning task. more than ten thousands
    remote sensing images are collected from Google Earth, Baidu Map, MapABC, Tianditu.
    The images are fixed to 224X224 pixels with various resolutions. The total number of
    remote sensing images are 10921, with five sentences descriptions per image.'
    """
    splits = ["train", "val", "test"]

    def __init__(
        self,
        root: str = ".data/rsicd",
        split: str = "train",
        transform: T.Compose = T.Compose([T.ToTensor()])
    ):
        assert split in self.splits
        self.root = root
        self.transform = transform
        self.captions = self.load_captions(os.path.join(root, "dataset_rsicd.json"), split)
        self.image_root = "RSICD_images"

    @staticmethod
    def load_captions(path: str, split: str) -> List[Dict]:
        with open(path) as f:
            captions = json.load(f)["images"]
        return [c for c in captions if c["split"] == split]

    def __len__(self) -> int:
        return len(self.captions)

    def __getitem__(self, idx: int) -> Dict:
        captions = self.captions[idx]
        path = os.path.join(self.root, self.image_root, captions["filename"])
        x = Image.open(path).convert("RGB")
        x = self.transform(x)
        sentences = [sentence["raw"] for sentence in captions["sentences"]]
        return dict(x=x, captions=sentences)



from typing import Tuple

import h5py
import torch
import torchvision.transforms as T


class SAT(torch.utils.data.Dataset):
    """ Base SAT dataset """
    splits = ["train", "test"]

    def __init__(
        self,
        root: str = "",
        split: str = "train",
        transform: T.Compose = T.Compose([T.ToTensor()])
    ):
        assert split in self.splits
        self.root = root
        self.split = split
        self.transform = transform

    def __len__(self) -> int:
        with h5py.File(self.root, "r") as f:
            length = f[f"{self.split}_y"].shape[0]
        return length

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, torch.Tensor]:
        with h5py.File(self.root, "r") as f:
            x = f[f"{self.split}_x"][idx]
            y = f[f"{self.split}_y"][idx]
        x = self.transform(x)
        return x, y


class SAT4(SAT):
    """ SAT-4 land cover classification dataset from "DeepSat - A Learning framework
    for Satellite Imagery", Basu et al (2015)
    https://arxiv.org/abs/1509.03602

    'SAT-4 consists of a total of 500,000 image patches covering four broad land cover classes.
    These include - barren land, trees, grassland and a class that consists of all land cover
    classes other than the above three. 400,000 patches (comprising of four-fifths of the total
    dataset) were chosen for training and the remaining 100,000 (one-fifths) were chosen as the testing
    dataset. We ensured that the training and test datasets belong to disjoint set of image tiles.
    Each image patch is size normalized to 28x28 pixels. Once generated, both the training and testing
    datasets were randomized using a pseudo-random number generator.'
    """
    classes = [
        "barren land",
        "trees",
        "grassland",
        "other"
    ]

    def __init__(self, root: str = ".data/sat/sat4.h5", *args, **kwargs):
        super().__init__(root, *args, **kwargs)


class SAT6(SAT):
    """ SAT-6 land cover classification dataset from "DeepSat - A Learning framework
    for Satellite Imagery", Basu et al (2015)
    https://arxiv.org/abs/1509.03602

    'SAT-6 consists of a total of 405,000 image patches each of size 28x28 and covering 6
    landcover classes - barren land, trees, grassland, roads, buildings and water bodies.
    324,000 images (comprising of four-fifths of the total dataset) were chosen as the training
    dataset and 81,000 (one fifths) were chosen as the testing dataset. Similar to SAT-4,
    the training and test sets were selected from disjoint NAIP tiles. Once generated, the
    images in the dataset were randomized in the same way as that for SAT-4. The specifications
    for the various landcover classes of SAT-4 and SAT-6 were adopted from those used in the
    National Land Cover Data (NLCD) algorithm.'
    """
    classes = [
        "barren land",
        "trees",
        "grassland",
        "roads",
        "buildings",
        "water"
    ]

    def __init__(self, root: str = ".data/sat/sat6.h5", *args, **kwargs):
        super().__init__(root, *args, **kwargs)



import os
from glob import glob
from typing import List, Dict

import torch
import tifffile

from torchrs.transforms import Compose, ToTensor


class HRSCD(torch.utils.data.Dataset):
    """ The High Resolution Semantic Change Detection (HRSCD) dataset from 'Multitask Learning
    for Large-scale Semantic Change Detection', Daudt at al. (2018)
    https://arxiv.org/abs/1810.08452

    'This dataset contains 291 coregistered image pairs of RGB aerial images from IGS's
    BD ORTHO database. Pixel-level change and land cover annotations are provided, generated
    by rasterizing Urban Atlas 2006, Urban Atlas 2012, and Urban Atlas Change 2006-2012 maps.'
    """
    classes = [
        "No Information",
        "Artificial surfaces",
        "Agricultural areas",
        "Forests",
        "Wetlands",
        "Water"
    ]

    def __init__(
        self,
        root: str = ".data/HRSCD",
        transform: Compose = Compose([ToTensor()])
    ):
        self.root = root
        self.transform = transform
        self.files = self.load_files(root)

    @staticmethod
    def load_files(root: str) -> List[Dict]:
        images1 = sorted(glob(os.path.join(root, "images", "2006", "**", "*.tif"), recursive=True))
        images2 = sorted(glob(os.path.join(root, "images", "2012", "**", "*.tif"), recursive=True))
        lcs1 = sorted(glob(os.path.join(root, "labels", "2006", "**", "*.tif"), recursive=True))
        lcs2 = sorted(glob(os.path.join(root, "labels", "2012", "**", "*.tif"), recursive=True))
        changes = sorted(glob(os.path.join(root, "labels", "change", "**", "*.tif"), recursive=True))
        files = []
        for image1, image2, lc1, lc2, change in zip(images1, images2, lcs1, lcs2, changes):
            region = image1.split(os.sep)[-2]
            files.append(dict(image1=image1, image2=image2, lc1=lc1, lc2=lc2, mask=change, region=region))
        return files

    def __len__(self) -> int:
        return len(self.files)

    def __getitem__(self, idx: int) -> Dict:
        """ Returns a dict containing x, land cover mask, change mask
        x: (2, 3, 1000, 1000)
        lc: (2, 1000, 1000)
        mask: (1, 1000, 1000)
        """
        files = self.files[idx]
        image1 = tifffile.imread(files["image1"])
        image2 = tifffile.imread(files["image2"])
        lc1 = tifffile.imread(files["lc1"])
        lc2 = tifffile.imread(files["lc2"])
        mask = tifffile.imread(files["mask"])
        image1, image2, lc1, lc2, mask = self.transform([image1, image2, lc1, lc2, mask])
        x = torch.stack([image1, image2], dim=0)
        lc = torch.cat([lc1, lc2], dim=0)
        return dict(x=x, lc=lc, mask=mask)



import os
import json
from glob import glob
from typing import List, Dict, Tuple

import torch
import numpy as np
from PIL import Image

from torchrs.transforms import Compose, ToTensor


def sort(x):
    x = os.path.basename(x)
    x = os.path.splitext(x)[0]
    return int(x)


class RSVQA(torch.utils.data.Dataset):
    """ Base RSVQA dataset """
    splits = ["train", "val", "test"]
    prefix = ""
    image_root = ""

    def __init__(
        self,
        root: str = "",
        split: str = "train",
        image_transform: Compose = Compose([ToTensor()]),
        text_transform: Compose = Compose([])
    ):
        assert split in self.splits
        self.root = root
        self.split = split
        self.image_transform = image_transform
        self.text_transform = text_transform
        self.image_root = os.path.join(root, self.image_root)
        self.ids, self.paths, self.images, self.questions, self.answers = self.load_files(
            self.root, self.image_root, self.split, self.prefix
        )

    @staticmethod
    def load_files(root: str, image_root: str, split: str, prefix: str) -> Tuple[List[int], List[str], List[Dict], List[Dict], List[Dict]]:
        paths = glob(os.path.join(image_root, "*.tif"))
        paths = sorted(paths, key=sort)
        with open(os.path.join(root, f"{prefix}_split_{split}_questions.json")) as f:
            questions = json.load(f)["questions"]
        with open(os.path.join(root, f"{prefix}_split_{split}_answers.json")) as f:
            answers = json.load(f)["answers"]
        with open(os.path.join(root, f"{prefix}_split_{split}_images.json")) as f:
            images = json.load(f)["images"]
        ids = [x["id"] for x in images if x["active"]]
        return ids, paths, images, questions, answers

    def __len__(self) -> int:
        return len(self.ids)

    def __getitem__(self, idx: int) -> Dict:
        """ Returns a dict containing x, questions, answers, q/a category
        x: (3, h, w)
        questions: List[str]
        answers: List[str]
        types: List[str]
        """
        id = self.ids[idx]
        x = np.array(Image.open(os.path.join(self.image_root, f"{id}.tif")))
        x = self.image_transform(x)
        questions = [self.questions[i] for i in self.images[id]["questions_ids"]]
        answers = [self.answers[q["answers_ids"][0]]["answer"] for q in questions]
        types = [q["type"] for q in questions]
        questions = [q["question"] for q in questions]
        questions = self.text_transform(questions)
        answers = self.text_transform(answers)
        output = dict(x=x, questions=questions, answers=answers, types=types)
        return output


class RSVQALR(RSVQA):
    """Remote Sensing Visual Question Answering Low Resolution (RSVQA LR) dataset from
    'RSVQA: Visual Question Answering for Remote Sensing Data', Lobry et al (2020)
    https://arxiv.org/abs/2003.07333

    'This dataset is based on Sentinel-2 images acquired over the Netherlands. Sentinel-2 satellites provide 10m resolution
    (for the visible bands used in this dataset) images with frequent updates (around 5 days) at a global scale. These images
    are openly available through ESA’s Copernicus Open Access Hub. To generate the dataset, we selected 9 Sentinel-2 tiles
    covering the Netherlands with a low cloud cover (selected tiles are shown in Figure 3). These tiles were divided in 772
    images of size 256x256 (covering 6.55km^2) retaining the RGB bands. From these, we constructed 770,232 questions and
    answers following the methodology presented in subsection II-A. We split the data in a training set (77.8% of the original tiles),
    a validation set (11.1%) and a test set (11.1%) at the tile level (the spatial split is shown in Figure 3). This allows
    to limit spatial correlation between the different splits.'
    """
    image_root = "Images_LR"
    prefix = "LR"

    def __init__(self, root: str = ".data/RSVQA_LR", *args, **kwargs):
        super().__init__(root, *args, **kwargs)


class RSVQAHR(RSVQA):
    """Remote Sensing Visual Question Answering High Resolution (RSVQA HR) dataset from
    'RSVQA: Visual Question Answering for Remote Sensing Data', Lobry et al (2020)
    https://arxiv.org/abs/2003.07333

    'This dataset uses 15cm resolution aerial RGB images extracted from the High Resolution
    Orthoimagery (HRO) data collection of the USGS. This collection covers most urban areas of the
    USA, along with a few areas of interest (e.g. national parks). For most areas covered by the dataset,
    only one tile is available with acquisition dates ranging from year 2000 to 2016, with various sensors.
    The tiles are openly accessible through USGS' EarthExplorer tool.

    From this collection, we extracted 161 tiles belonging to the North-East coast of the USA
    that were split into 100659 images of size 512x512 (each covering 5898m^2).We constructed 100,660,316
    questions and answers following the methodology presented in subsection II-A. We split the data in
    a training set (61.5% of the tiles), a validation set (11.2%), and test sets (20.5% for test set 1,
    6.8% for test set 2). As it can be seen in Figure 4, test set 1 covers similar regions as the training
    and validation sets, while test set 2 covers the city of Philadelphia, which is not seen during the
    training. Note that this second test set also uses another sensor (marked as unknown on the USGS
    data catalog), not seen during training.
    """
    image_root = "Data"
    prefix = "USGS"

    def __init__(self, root: str = ".data/RSVQA_HR", *args, **kwargs):
        super().__init__(root, *args, **kwargs)


class RSVQAxBEN(RSVQA):
    """Remote Sensing Visual Question Answering BigEarthNet (RSVQAxBEN) dataset from
    'RSVQA Meets BigEarthNet: A New, Large-Scale, Visual Question Answering Dataset for Remote Sensing', Lobry et al (2021)
    https://rsvqa.sylvainlobry.com/IGARSS21.pdf

    'We introduce a new dataset to tackle the task of visual question answering on remote
    sensing images: this largescale, open access dataset extracts image/question/answer triplets
    from the BigEarthNet dataset. This new dataset contains close to 15 millions samples and is openly
    available.'
    """
    image_root = "Images"
    prefix = "RSVQAxBEN"

    def __init__(self, root: str = ".data/rsvqaxben", *args, **kwargs):
        super().__init__(root, *args, **kwargs)



import os
import re
from glob import glob
from typing import List, Dict

import torch
import numpy as np
from PIL import Image

from torchrs.transforms import Compose, ToTensor


class InriaAIL(torch.utils.data.Dataset):
    """ Inria Aerial Image Labeling dataset from 'Can semantic labeling methods
    generalize to any city? the inria aerial image labeling benchmark', Maggiori et al. (2017)
    https://ieeexplore.ieee.org/document/8127684

    'The training set contains 180 color image tiles of size 5000x5000, covering a surface of 1500mx1500m
    each (at a 30 cm resolution). There are 36 tiles for each of the following regions (Austin, Chicago, Kitsap County, Western Tyrol, Vienna)
    The format is GeoTIFF. Files are named by a prefix associated to the region (e.g., austin- or vienna-)
    followed by the tile number (1-36). The reference data is in a different folder and the file names
    correspond exactly to those of the color images. In the case of the reference data, the tiles are
    single-channel images with values 255 for the building class and 0 for the not building class.'
    """
    splits = ["train", "test"]

    def __init__(
        self,
        root: str = ".data/AerialImageDataset",
        split: str = "train",
        transform: Compose = Compose([ToTensor()]),
    ):
        self.split = split
        self.transform = transform
        self.images = self.load_images(root, split)
        self.regions = sorted(list(set(image["region"] for image in self.images)))

    @staticmethod
    def load_images(path: str, split: str) -> List[Dict]:
        images = sorted(glob(os.path.join(path, split, "images", "*.tif")))
        pattern = re.compile("[a-zA-Z]+")
        regions = [re.findall(pattern, os.path.basename(image))[0] for image in images]

        if split == "train":
            targets = sorted(glob(os.path.join(path, split, "gt", "*.tif")))
        else:
            targets = [None] * len(images)

        files = [
            dict(image=image, target=target, region=region)
            for image, target, region in zip(images, targets, regions)
        ]
        return files

    def __len__(self) -> int:
        return len(self.images)

    def __getitem__(self, idx: int) -> Dict:
        image_path, target_path = self.images[idx]["image"], self.images[idx]["target"]
        x = np.array(Image.open(image_path))

        if self.split == "train":
            y = np.array(Image.open(target_path))
            y = np.clip(y, a_min=0, a_max=1)
            x, y = self.transform([x, y])
            output = dict(x=x, mask=y, region=self.images[idx]["region"])
        else:
            x = self.transform(x)
            output = dict(x=x)

        return output



from typing import Dict

import torch
import torch.nn as nn
import torchmetrics
import pytorch_lightning as pl

from torchrs.models import RAMS


class RAMSModule(pl.LightningModule):

    def __init__(
        self,
        scale_factor: int = 3,
        t: int = 9,
        c: int = 1,
        num_feature_attn_blocks: int = 12,
        loss_fn: nn.Module = nn.MSELoss(),
        opt: torch.optim.Optimizer = torch.optim.Adam,
        lr: float = 3E-4
    ):
        super(RAMSModule, self).__init__()
        self.loss_fn = loss_fn
        self.opt = opt
        self.lr = lr
        self.model = RAMS(scale_factor, t, c, num_feature_attn_blocks)

        metrics = torchmetrics.MetricCollection([
            torchmetrics.MeanSquaredError(),
            torchmetrics.MeanAbsoluteError(),
            torchmetrics.MeanAbsolutePercentageError(),
            torchmetrics.PSNR(),
            torchmetrics.SSIM()
        ])
        self.train_metrics = metrics.clone(prefix='train_')
        self.val_metrics = metrics.clone(prefix='val_')
        self.test_metrics = metrics.clone(prefix='test_')

    def configure_optimizers(self):
        return self.opt(self.parameters(), lr=self.lr)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)

    def training_step(self, batch: Dict, batch_idx: int):
        lr, hr = batch["lr"], batch["hr"]
        sr = self(lr)
        loss = self.loss_fn(sr, hr)
        metrics = self.train_metrics(sr.to(torch.float32), hr)
        metrics["train_loss"] = loss
        self.log_dict(metrics)
        return loss

    def validation_step(self, batch: Dict, batch_idx: int):
        lr, hr = batch["lr"], batch["hr"]
        sr = self(lr)
        loss = self.loss_fn(sr, hr)
        metrics = self.val_metrics(sr.to(torch.float32), hr)
        metrics["val_loss"] = loss
        self.log_dict(metrics)

    def test_step(self, batch: Dict, batch_idx: int):
        lr, hr = batch["lr"], batch["hr"]
        sr = self(lr)
        loss = self.loss_fn(sr, hr)
        metrics = self.test_metrics(sr.to(torch.float32), hr)
        metrics["test_loss"] = loss
        self.log_dict(metrics)



import os
from glob import glob
from typing import List, Dict

import torch
import numpy as np
from PIL import Image

from torchrs.transforms import Compose, ToTensor


class GID15(torch.utils.data.Dataset):
    """ Gaofen Image Dataset (GID-15) from 'Land-Cover Classification with High-Resolution
    Remote Sensing Images Using Transferable Deep Models', Tong et al. (2018)
    https://arxiv.org/abs/1807.05713

    'We construct a new large-scale land-cover dataset with Gaofen-2 (GF-2) satellite
    images. This new dataset, which is named as Gaofen Image Dataset with 15 categories
    (GID-15), has superiorities over the existing land-cover dataset because of its
    large coverage, wide distribution, and high spatial resolution. The large-scale
    remote sensing semantic segmentation set contains 150 pixel-level annotated GF-2
    images, which is labeled in 15 categories.'
    """
    classes = [
        "background",
        "industrial_land",
        "urban_residential",
        "rural_residential",
        "traffic_land",
        "paddy_field",
        "irrigated_land",
        "dry_cropland",
        "garden_plot",
        "arbor_woodland",
        "shrub_land",
        "natural_grassland",
        "artificial_grassland",
        "river",
        "lake",
        "pond"
    ]
    splits = ["train", "val", "test"]

    def __init__(
        self,
        root: str = ".data/gid-15",
        split: str = "train",
        transform: Compose = Compose([ToTensor()]),
    ):
        self.split = split
        self.transform = transform
        self.images = self.load_images(os.path.join(root, "GID"), split)

    @staticmethod
    def load_images(path: str, split: str) -> List[Dict]:
        images = sorted(glob(os.path.join(path, "img_dir", split, "*.tif")))
        if split in ["train", "val"]:
            masks = [
                image.replace("img_dir", "ann_dir").replace(".tif", "_15label.png")
                for image in images
            ]
        else:
            masks = [None] * len(images)

        files = [dict(image=image, mask=mask) for image, mask in zip(images, masks)]
        return files

    def __len__(self) -> int:
        return len(self.images)

    def __getitem__(self, idx: int) -> Dict:
        image_path, mask_path = self.images[idx]["image"], self.images[idx]["mask"]
        x = np.array(Image.open(image_path))

        if self.split in ["train", "val"]:
            y = np.array(Image.open(mask_path))
            x, y = self.transform([x, y])
            output = dict(x=x, mask=y)
        else:
            x = self.transform(x)
            output = dict(x=x)

        return output



from .base import BaseDataModule
from .probav import PROBAVDataModule
from .etci2021 import ETCI2021DataModule
from .rsvqa import RSVQALRDataModule, RSVQAHRDataModule, RSVQAxBENDataModule
from .eurosat import EuroSATRGBDataModule, EuroSATMSDataModule
from .resisc45 import RESISC45DataModule
from .rsicd import RSICDDataModule
from .oscd import OSCDDataModule
from .s2looking import S2LookingDataModule
from .levircd import LEVIRCDPlusDataModule
from .fair1m import FAIR1MDataModule
from .sydney_captions import SydneyCaptionsDataModule
from .ucm_captions import UCMCaptionsDataModule
from .s2mtcp import S2MTCPDataModule
from .advance import ADVANCEDataModule
from .sat import SAT4DataModule, SAT6DataModule
from .hrscd import HRSCDDataModule
from .inria_ail import InriaAILDataModule
from .tiselac import TiselacDataModule
from .gid15 import GID15DataModule
from .zuericrop import ZueriCropDataModule
from .aid import AIDDataModule
from .dubai_segmentation import DubaiSegmentationDataModule
from .hkh_glacier import HKHGlacierMappingDataModule
from .ucm import UCMDataModule
from .patternnet import PatternNetDataModule
from .whu_rs19 import WHURS19DataModule
from .rsscn7 import RSSCN7DataModule
from .brazilian_coffee import BrazilianCoffeeScenesDataModule


__all__ = [
    "BaseDataModule", "PROBAVDataModule", "ETCI2021DataModule", "RSVQALRDataModule",
    "RSVQAxBENDataModule", "EuroSATRGBDataModule", "EuroSATMSDataModule", "RESISC45DataModule",
    "RSICDDataModule", "OSCDDataModule", "S2LookingDataModule", "LEVIRCDPlusDataModule",
    "FAIR1MDataModule", "SydneyCaptionsDataModule", "UCMCaptionsDataModule", "S2MTCPDataModule",
    "ADVANCEDataModule", "SAT4DataModule", "SAT6DataModule", "HRSCDDataModule", "InriaAILDataModule",
    "TiselacDataModule", "GID15DataModule", "ZueriCropDataModule", "AIDDataModule",
    "DubaiSegmentationDataModule", "HKHGlacierMappingDataModule", "UCMDataModule", "PatternNetDataModule",
    "RSVQAHRDataModule", "WHURS19DataModule", "RSSCN7DataModule", "BrazilianCoffeeScenesDataModule"
]



import torchvision.transforms as T
from torchvision.datasets import ImageFolder


class RESISC45(ImageFolder):
    """ Image Scene Classification dataset from 'Remote Sensing Image
    Scene Classification: Benchmark and State of the Art', Cheng at al. (2017)
    https://arxiv.org/abs/1703.00121

    'We propose a large-scale dataset, termed "NWPU-RESISC45", which is a publicly
    available benchmark for REmote Sensing Image Scene Classification (RESISC), created
    by Northwestern Polytechnical University (NWPU). This dataset contains 31,500 images,
    covering 45 scene classes with 700 images in each class. The proposed NWPU-RESISC45 (i)
    is large-scale on the scene classes and the total image number, (ii) holds big variations
    in translation, spatial resolution, viewpoint, object pose, illumination, background, and
    occlusion, and (iii) has high within-class diversity and between-class similarity.'
    """
    def __init__(
        self,
        root: str = ".data/NWPU-RESISC45",
        transform: T.Compose = T.Compose([T.ToTensor()])
    ):
        super().__init__(
            root=root,
            transform=transform
        )



from setuptools import setup, find_packages
from setuptools.config import read_configuration

extras = {
    "train": ["pytorch-lightning>=1.4.0", "torchmetrics>=0.4.1"],
}
install_requires = [
    "torch>=1.9.0",
    "torchvision>=0.10.0",
    "torchaudio>=0.9.0",
    "einops>=0.3.0",
    "numpy>=1.21.0",
    "pillow>=8.3.1",
    "tifffile>=2021.7.2",
    "h5py>=3.3.0",
    "imagecodecs>=2021.7.30"
]
setup_requires = ["pytest-runner"]
tests_require = ["pytest", "pytest-cov", "mock", "mypy", "black", "pylint"]

cfg = read_configuration("setup.cfg")

setup(
    download_url='{}/archive/{}.tar.gz'.format(cfg["metadata"]["url"], cfg["metadata"]["version"]),
    project_urls={"Bug Tracker": cfg["metadata"]["url"] + "/issues"},
    install_requires=install_requires,
    extras_require=extras,
    setup_requires=setup_requires,
    tests_require=tests_require,
    packages=find_packages(),
    python_requires=">=3.7",
)



import os
from glob import glob
from xml.etree import ElementTree
from typing import List, Dict

import torch
import numpy as np
import torchvision.transforms as T
from PIL import Image


def parse_pascal_voc(path: str) -> Dict:
    et = ElementTree.parse(path)
    element = et.getroot()
    image = element.find("source").find("filename").text
    classes, points = [], []
    for obj in element.find("objects").findall("object"):
        obj_points = [p.text.split(",") for p in obj.find("points").findall("point")]
        obj_points = [(float(p1), float(p2)) for p1, p2 in obj_points]
        cls = obj.find("possibleresult").find("name").text
        classes.append(cls)
        points.append(obj_points)
    return dict(image=image, points=points, classes=classes)


class FAIR1M(torch.utils.data.Dataset):
    """ FAIR1M dataset from 'FAIR1M: A Benchmark Dataset
    for Fine-grained Object Recognition in High-Resolution Remote Sensing Imagery', Sun at al. (2021)
    https://arxiv.org/abs/2103.05569

    'We propose a novel benchmark dataset with more than 1 million instances and more
    than 15,000 images for Fine-grained object recognition in high-resolution remote
    sensing imagery which is named as FAIR1M. We collected remote sensing images with
    a resolution of 0.3m to 0.8m from different platforms, which are spread across many
    countries and regions. All objects in the FAIR1M dataset are annotated with respect
    to 5 categories and 37 sub-categories by oriented bounding boxes.'
    """
    classes = {
        "Passenger Ship":   {"id": 0, "category": "Ship"},
        "Motorboat":        {"id": 1, "category": "Ship"},
        "Fishing Boat":     {"id": 2, "category": "Ship"},
        "Tugboat":          {"id": 3, "category": "Ship"},
        "other-ship":       {"id": 4, "category": "Ship"},
        "Engineering Ship": {"id": 5, "category": "Ship"},
        "Liquid Cargo Ship": {"id": 6, "category": "Ship"},
        "Dry Cargo Ship":   {"id": 7, "category": "Ship"},
        "Warship":          {"id": 8, "category": "Ship"},
        "Small Car":        {"id": 9, "category": "Vehicle"},
        "Bus":              {"id": 10, "category": "Vehicle"},
        "Cargo Truck":      {"id": 11, "category": "Vehicle"},
        "Dump Truck":       {"id": 12, "category": "Vehicle"},
        "other-vehicle":    {"id": 13, "category": "Vehicle"},
        "Van":              {"id": 14, "category": "Vehicle"},
        "Trailer":          {"id": 15, "category": "Vehicle"},
        "Tractor":          {"id": 16, "category": "Vehicle"},
        "Excavator":        {"id": 17, "category": "Vehicle"},
        "Truck Tractor":    {"id": 18, "category": "Vehicle"},
        "Boeing737":        {"id": 19, "category": "Airplane"},
        "Boeing747":        {"id": 20, "category": "Airplane"},
        "Boeing777":        {"id": 21, "category": "Airplane"},
        "Boeing787":        {"id": 22, "category": "Airplane"},
        "ARJ21":            {"id": 23, "category": "Airplane"},
        "C919":             {"id": 24, "category": "Airplane"},
        "A220":             {"id": 25, "category": "Airplane"},
        "A321":             {"id": 26, "category": "Airplane"},
        "A330":             {"id": 27, "category": "Airplane"},
        "A350":             {"id": 28, "category": "Airplane"},
        "other-airplane":   {"id": 29, "category": "Airplane"},
        "Baseball Field":   {"id": 30, "category": "Court"},
        "Basketball Court": {"id": 31, "category": "Court"},
        "Football Field":   {"id": 32, "category": "Court"},
        "Tennis Court":     {"id": 33, "category": "Court"},
        "Roundabout":       {"id": 34, "category": "Road"},
        "Intersection":     {"id": 35, "category": "Road"},
        "Bridge":           {"id": 36, "category": "Road"}
    }

    def __init__(
        self,
        root: str = ".data/fair1m",
        transform: T.Compose = T.Compose([T.ToTensor()]),
    ):
        split = "train"
        self.root = root
        self.image_root = os.path.join(root, split, "part1", "images")
        self.transform = transform
        self.images = self.load_files(root, split)
        self.idx2cls = {i: c for i, c in enumerate(self.classes)}
        self.cls2idx = {c: i for i, c in self.idx2cls.items()}

    @staticmethod
    def load_files(root: str, split: str) -> List[Dict]:
        files = sorted(glob(os.path.join(root, split, "part1", "labelXmls", "*.xml")))
        return [parse_pascal_voc(f) for f in files]

    def __len__(self) -> int:
        return len(self.images)

    def __getitem__(self, idx: int) -> Dict:
        """ Returns a dict containing x, y, points where points is the x,y coords of the rotated bbox
        x: (3, h, w)
        y: (N,)
        points: (N, 5, 2)
        """
        image = self.images[idx]
        x = np.array(Image.open(os.path.join(self.image_root, image["image"])))
        x = x[..., :3]
        x = self.transform(x)
        y = torch.tensor([self.cls2idx[c] for c in image["classes"]])
        points = torch.tensor(image["points"])
        return dict(x=x, y=y, points=points)



import os
from typing import Tuple

import torch
import numpy as np
from einops import rearrange

from torchrs.transforms import Compose, ToTensor


class Tiselac(torch.utils.data.Dataset):
    """ TiSeLac dataset from the Time Series Land Cover Classification Challenge (2017)
    https://sites.google.com/site/dinoienco/tiselac-time-series-land-cover-classification-challenge

    'A MSTC Land Cover classification problem for data taken from the Reunion island.
    A case is a pixel. Measurements are taken over 23 time points (days), with
    10 dimensions: 7 surface reflectances (Ultra Blue, Blue, Green, Red, NIR, SWIR1 and SWIR2)
    plus 3 indices (NDVI, NDWI and BI). Class values relate to one of 9 land cover types class values.'
    """
    classes = [
        "Urban Areas",
        "Other built-up surfaces",
        "Forests",
        "Sparse Vegetation",
        "Rocks and bare soil",
        "Grassland",
        "Sugarcane crops",
        "Other crops",
        "Water"
    ]
    splits = ["train", "test"]

    def __init__(
        self,
        root: str = ".data/tiselac",
        split: str = "train",
        transform: Compose = Compose([ToTensor(permute_dims=False)])
    ):
        assert split in self.splits
        self.root = root
        self.transform = transform
        self.series, self.labels = self.load_file(root, split)

    @staticmethod
    def load_file(path: str, split: str) -> Tuple[np.ndarray, np.ndarray]:
        x = np.loadtxt(os.path.join(path, f"{split}.txt"), dtype=np.int16, delimiter=",")
        y = np.loadtxt(os.path.join(path, f"{split}_labels.txt"), dtype=np.uint8)
        x = rearrange(x, "n (t c) -> n t c", c=10)
        return x, y

    def __len__(self) -> int:
        return len(self.series)

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, torch.Tensor]:
        x, y = self.series[idx], self.labels[idx] - 1
        x, y = self.transform(x).squeeze(dim=0), torch.tensor(y).to(torch.long)
        return x, y



import os
from glob import glob
from typing import List, Dict

import torch
import numpy as np

from torchrs.transforms import Compose, ToTensor


class HKHGlacierMapping(torch.utils.data.Dataset):
    """ Hindu Kush Himalayas (HKH) Glacier Mapping dataset
    https://lila.science/datasets/hkh-glacier-mapping

    'We also provide 14190 numpy patches. The numpy patches are all of size 512x512x15 and
    corresponding 512x512x2 pixel-wise mask labels; the two channels in the pixel-wise masks
    correspond to clean-iced and debris-covered glaciers. Patches' geolocation information,
    time stamps, source Landsat IDs, and glacier density are available in a geojson metadata file.'

    """
    bands = [
        "LE7 B1 (blue)",
        "LE7 B2 (green)",
        "LE7 B3 (red)",
        "LE7 B4 (near infrared)",
        "LE7 B5 (shortwave infrared 1)",
        "LE7 B6_VCID_1 (low-gain thermal infrared)",
        "LE7 B6_VCID_2 (high-gain thermal infrared)",
        "LE7 B7 (shortwave infrared 2)",
        "LE7 B8 (panchromatic)",
        "LE7 BQA (quality bitmask)",
        "NDVI (vegetation index)",
        "NDSI (snow index)",
        "NDWI (water index)",
        "SRTM 90 elevation",
        "SRTM 90 slope"
    ]

    def __init__(
        self,
        root: str = ".data/hkh_glacier_mapping",
        transform: Compose = Compose([ToTensor()]),
    ):
        self.transform = transform
        self.images = self.load_images(root)

    @staticmethod
    def load_images(path: str) -> List[Dict]:
        images = sorted(glob(os.path.join(path, "images", "*.npy")))
        masks = sorted(glob(os.path.join(path, "masks", "*.npy")))
        return [dict(image=image, mask=mask) for image, mask in zip(images, masks)]

    def __len__(self) -> int:
        return len(self.images)

    def __getitem__(self, idx: int) -> Dict[str, torch.Tensor]:
        image_path, target_path = self.images[idx]["image"], self.images[idx]["mask"]
        x, y = np.load(image_path), np.load(target_path)
        y0, y1 = y[..., 0], y[..., 1]
        x, y0, y1 = self.transform([x, y0, y1])
        return dict(x=x, clean_ice_mask=y0, debris_covered_mask=y1)



import torchvision.transforms as T
from torchvision.datasets import ImageFolder


class RSSCN7(ImageFolder):
    """ RSSCN7 dataset from 'Deep Learning Based Feature Selection for
    Remote Sensing Scene Classification', Zou at al. (2015)
    https://ieeexplore.ieee.org/abstract/document/7272047

    'The data set RSSCN7 contains 2800 remote sensing scene images, which
    are from seven typical scene categories, namely, the grassland, forest,
    farmland, parking lot, residential region, industrial region, and river
    and lake. For each category, there are 400 images collected from the
    Google Earth, which are sampled on four different scales with 100 images
    per scale. Each image has a size of 400x400 pixels. This data set is
    rather challenging due to the wide diversity of the scene images that
    are captured under changing seasons and varying weathers and sampled
    on different scales'
    """
    def __init__(
        self,
        root: str = ".data/RSSCN7",
        transform: T.Compose = T.Compose([T.ToTensor()])
    ):
        super().__init__(
            root=root,
            transform=transform
        )



import os
from glob import glob
from typing import List, Dict

import torch
import numpy as np

from torchrs.transforms import Compose, ToTensor


class S2MTCP(torch.utils.data.Dataset):
    """ Sentinel-2 Multitemporal Cities Pairs (S2MTCP) dataset from 'Self-supervised
    pre-training enhances change detection in Sentinel-2 imagery', Leenstra at al. (2021)
    https://arxiv.org/abs/2101.08122

    'A dataset of S2 level 1C image pairs was created ... Image pairs are selected randomly
    from available S2 images of each location with less than one percent cloud cover. Bands
    with a spatial resolution smaller than 10 m are resampled to 10 m and images are cropped
    to approximately 600x600 pixels centered on the selected coordinates ... The S2MTCP dataset
    contains N = 1520 image pairs, spread over all inhabited continents, with the highest
    concentration of image pairs in North-America, Europe and Asia'

    Note this dataset doesn't contain change masks as it was created for use in self-supervised pretraining
    """
    def __init__(
        self,
        root: str = ".data/s2mtcp",
        transform: Compose = Compose([ToTensor()]),
    ):
        self.root = os.path.join(root, "data_S21C")
        self.transform = transform
        self.files = self.load_files(self.root)

    @staticmethod
    def load_files(root: str) -> List[Dict]:
        files = glob(os.path.join(root, "*.npy"))
        files = [os.path.basename(f).split("_")[0] for f in files]
        files = sorted(set(files), key=int)
        files = [dict(image1=f"{num}_a.npy", image2=f"{num}_b.npy") for num in files]
        return files

    def __len__(self) -> int:
        return len(self.files)

    def __getitem__(self, idx: int) -> torch.Tensor:
        """ Returns x: (2, 14, h, w) """
        files = self.files[idx]
        image1 = np.load(os.path.join(self.root, files["image1"]))
        image2 = np.load(os.path.join(self.root, files["image2"]))
        image1, image2 = self.transform([image1, image2])
        x = torch.stack([image1, image2], dim=0)
        return x
